# mypy: ignore-errors
import math
import numbers
import warnings

import casadi
import numpy as np
import pybammsolvers.idaklu as idaklu

import pybamm


class IDAKLUSolver(pybamm.BaseSolver):
    """
    Solve a discretised model, using sundials with the KLU sparse linear solver.

    Parameters
    ----------
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-4).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not (default is 0).
    on_extrapolation : str, optional
        What to do if the solver is extrapolating. Options are "warn", "error", or "ignore".
        Default is "warn".
    on_failure : str, optional
        What to do if a solver error flag occurs. Options are "warn", "error", or "ignore".
        Default is "raise".
    output_variables : list[str], optional
        List of variables to calculate and return. If none are specified then
        the complete state vector is returned (can be very large) (default is [])
    options: dict, optional
        Addititional options to pass to the solver, by default:

        .. code-block:: python

            options = {
                # Print statistics of the solver after every solve
                "print_stats": False,
                # Number of threads available for OpenMP (must be greater than or equal to `num_solvers`)
                "num_threads": 1,
                # Number of solvers to use in parallel (for solving multiple sets of input parameters in parallel)
                "num_solvers": num_threads,
                ## Linear solver interface
                # name of sundials linear solver to use options are: "SUNLinSol_KLU",
                # "SUNLinSol_Dense", "SUNLinSol_Band", "SUNLinSol_SPBCGS",
                # "SUNLinSol_SPFGMR", "SUNLinSol_SPGMR", "SUNLinSol_SPTFQMR",
                "linear_solver": "SUNLinSol_KLU",
                # Jacobian form, can be "none", "dense",
                # "banded", "sparse", "matrix-free"
                "jacobian": "sparse",
                # Preconditioner for iterative solvers, can be "none", "BBDP"
                "preconditioner": "BBDP",
                # For iterative linear solver preconditioner, bandwidth of
                # approximate jacobian
                "precon_half_bandwidth": 5,
                # For iterative linear solver preconditioner, bandwidth of
                # approximate jacobian that is kept
                "precon_half_bandwidth_keep": 5,
                # For iterative linear solvers, max number of iterations
                "linsol_max_iterations": 5,
                # Ratio between linear and nonlinear tolerances
                "epsilon_linear_tolerance": 0.05,
                # Increment factor used in DQ Jacobian-vector product approximation
                "increment_factor": 1.0,
                # Enable or disable linear solution scaling
                "linear_solution_scaling": True,
                ## Main solver
                # Maximum order of the linear multistep method
                "max_order_bdf": 5,
                # Maximum number of steps to be taken by the solver in its attempt to
                # reach the next output time.
                # Note: this value differs from the IDA default of 500
                "max_num_steps": 100000,
                # Initial step size. The solver default is used if this is left at 0.0
                "dt_init": 0.0,
                # Minimum absolute step size. The solver default is used if this is
                # left at 0.0
                "dt_min": 0.0,
                # Maximum absolute step size. The solver default is used if this is
                # left at 0.0
                "dt_max": 0.0,
                # Maximum number of error test failures in attempting one step
                "max_error_test_failures": 10,
                # Maximum number of nonlinear solver iterations at one step
                # Note: this value differs from the IDA default of 4
                "max_nonlinear_iterations": 40,
                # Maximum number of nonlinear solver convergence failures at one step
                # Note: this value differs from the IDA default of 10
                "max_convergence_failures": 100,
                # Safety factor in the nonlinear convergence test
                "nonlinear_convergence_coefficient": 0.33,
                # Suppress algebraic variables from error test
                "suppress_algebraic_error": False,
                # Store Hermite interpolation data for the solution.
                # Note: this option is always disabled if output_variables are given
                # or if t_interp values are specified
                "hermite_interpolation": True,
                ## Initial conditions calculation
                # Positive constant in the Newton iteration convergence test within the
                # initial condition calculation
                "nonlinear_convergence_coefficient_ic": 0.0033,
                # Maximum number of steps allowed when `init_all_y_ic = False`
                # Note: this value differs from the IDA default of 5
                "max_num_steps_ic": 50,
                # Maximum number of the approximate Jacobian or preconditioner evaluations
                # allowed when the Newton iteration appears to be slowly converging
                # Note: this value differs from the IDA default of 4
                "max_num_jacobians_ic": 40,
                # Maximum number of Newton iterations allowed in any one attempt to solve
                # the initial conditions calculation problem
                # Note: this value differs from the IDA default of 10
                "max_num_iterations_ic": 100,
                # Maximum number of linesearch backtracks allowed in any Newton iteration,
                # when solving the initial conditions calculation problem
                "max_linesearch_backtracks_ic": 100,
                # Turn off linesearch
                "linesearch_off_ic": False,
                # How to calculate the initial conditions.
                # "True": calculate all y0 given ydot0
                # "False": calculate y_alg0 and ydot_diff0 given y_diff0
                "init_all_y_ic": False,
                # Calculate consistent initial conditions
                "calc_ic": True,
                ## Early termination
                # Maximum number of consecutive steps allowed without advancing
                # the solution time by at least `t_no_progress` seconds.
                # If set to 0, this feature is disabled.
                "num_steps_no_progress": 0,
                # Minimum required time advancement (in seconds) after
                # `num_steps_no_progress` consecutive steps.
                # If set to 0.0, this feature is disabled.
                "t_no_progress": 0.0,
            }

        Note: These options only have an effect if model.convert_to_format == 'casadi'


    """

    def __init__(
        self,
        rtol=1e-4,
        atol=1e-6,
        root_method="casadi",
        root_tol=1e-6,
        extrap_tol=None,
        on_extrapolation=None,
        output_variables=None,
        on_failure=None,
        options=None,
    ):
        # set default options,
        # (only if user does not supply)
        default_options = {
            "print_stats": False,
            "jacobian": "sparse",
            "preconditioner": "BBDP",
            "precon_half_bandwidth": 5,
            "precon_half_bandwidth_keep": 5,
            "num_threads": 1,
            "num_solvers": 1,
            "linear_solver": "SUNLinSol_KLU",
            "linsol_max_iterations": 5,
            "epsilon_linear_tolerance": 0.05,
            "increment_factor": 1.0,
            "linear_solution_scaling": True,
            "max_order_bdf": 5,
            "max_num_steps": 100000,
            "dt_init": 0.0,
            "dt_min": 0.0,
            "dt_max": 0.0,
            "max_error_test_failures": 10,
            "max_nonlinear_iterations": 40,
            "max_convergence_failures": 100,
            "nonlinear_convergence_coefficient": 0.33,
            "suppress_algebraic_error": False,
            "hermite_interpolation": True,
            "nonlinear_convergence_coefficient_ic": 0.0033,
            "max_num_steps_ic": 50,
            "max_num_jacobians_ic": 40,
            "max_num_iterations_ic": 100,
            "max_linesearch_backtracks_ic": 100,
            "linesearch_off_ic": False,
            "init_all_y_ic": False,
            "calc_ic": True,
            "num_steps_no_progress": 0,
            "t_no_progress": 0.0,
        }
        if options is None:
            options = default_options
        else:
            if "num_threads" in options and "num_solvers" not in options:
                options["num_solvers"] = options["num_threads"]
            for key, value in default_options.items():
                if key not in options:
                    options[key] = value
        self._options = options

        self.output_variables = [] if output_variables is None else output_variables

        super().__init__(
            method="ida",
            rtol=rtol,
            atol=atol,
            root_method=root_method,
            root_tol=root_tol,
            extrap_tol=extrap_tol,
            output_variables=output_variables,
            on_extrapolation=on_extrapolation,
            on_failure=on_failure,
        )
        self.name = "IDA KLU solver"
        self._supports_interp = True
        self._supports_t_eval_discontinuities = True

        pybamm.citations.register("Hindmarsh2000")
        pybamm.citations.register("Hindmarsh2005")

    def _check_atol_type(self, atol, size):
        """
        This method checks that the atol vector is of the right shape and
        type.

        Parameters
        ----------
        atol: double or np.array or list
            Absolute tolerances. If this is a vector then each entry corresponds to
            the absolute tolerance of one entry in the state vector.
        size: int
            The length of the atol vector
        """

        if isinstance(atol, float):
            atol = atol * np.ones(size)
        elif not isinstance(atol, np.ndarray):
            raise pybamm.SolverError(
                "Absolute tolerances must be a numpy array or float"
            )

        return atol

    def set_up(self, model, inputs=None, t_eval=None, ics_only=False):
        base_set_up_return = super().set_up(model, inputs, t_eval, ics_only)

        inputs_dict = inputs or {}
        # stack inputs
        if inputs_dict:
            arrays_to_stack = [np.array(x).reshape(-1, 1) for x in inputs_dict.values()]
            inputs = np.vstack(arrays_to_stack)
        else:
            inputs = np.array([[]])

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full()
        y0 = y0.flatten()

        if ics_only:
            return base_set_up_return

        if model.convert_to_format != "casadi":
            msg = "The python-idaklu solver has been deprecated."
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            raise pybamm.SolverError(
                f"Unsupported option for convert_to_format={model.convert_to_format}"
            )

        if self._options["jacobian"] == "dense":
            mass_matrix = casadi.DM(model.mass_matrix.entries.toarray())
        else:
            mass_matrix = casadi.DM(model.mass_matrix.entries)

        # construct residuals function by binding inputs
        # TODO: do we need densify here?
        rhs_algebraic = model.rhs_algebraic_eval

        if not model.use_jacobian:
            raise pybamm.SolverError("KLU requires the Jacobian")

        # need to provide jacobian_rhs_alg - cj * mass_matrix
        t_casadi = casadi.MX.sym("t")
        y_casadi = casadi.MX.sym("y", model.len_rhs_and_alg)
        cj_casadi = casadi.MX.sym("cj")
        p_casadi = {}
        for name, value in inputs_dict.items():
            if isinstance(value, numbers.Number):
                p_casadi[name] = casadi.MX.sym(name)
            else:
                p_casadi[name] = casadi.MX.sym(name, value.shape[0])
        p_casadi_stacked = casadi.vertcat(*[p for p in p_casadi.values()])

        jac_times_cjmass = casadi.Function(
            "jac_times_cjmass",
            [t_casadi, y_casadi, p_casadi_stacked, cj_casadi],
            [
                model.jac_rhs_algebraic_eval(t_casadi, y_casadi, p_casadi_stacked)
                - cj_casadi * mass_matrix
            ],
        )

        jac_times_cjmass_sparsity = jac_times_cjmass.sparsity_out(0)
        jac_bw_lower = jac_times_cjmass_sparsity.bw_lower()
        jac_bw_upper = jac_times_cjmass_sparsity.bw_upper()
        jac_times_cjmass_nnz = jac_times_cjmass_sparsity.nnz()
        jac_times_cjmass_colptrs = np.array(
            jac_times_cjmass_sparsity.colind(), dtype=np.int64
        )
        jac_times_cjmass_rowvals = np.array(
            jac_times_cjmass_sparsity.row(), dtype=np.int64
        )

        v_casadi = casadi.MX.sym("v", model.len_rhs_and_alg)

        jac_rhs_algebraic_action = model.jac_rhs_algebraic_action_eval

        # also need the action of the mass matrix on a vector
        mass_action = casadi.Function(
            "mass_action", [v_casadi], [casadi.densify(mass_matrix @ v_casadi)]
        )

        num_of_events = len(model.terminate_events_eval)

        # rootfn needs to return an array of length num_of_events
        rootfn = casadi.Function(
            "rootfn",
            [t_casadi, y_casadi, p_casadi_stacked],
            [
                casadi.vertcat(
                    *[
                        event(t_casadi, y_casadi, p_casadi_stacked)
                        for event in model.terminate_events_eval
                    ]
                )
            ],
        )

        # get ids of rhs and algebraic variables
        rhs_ids = np.ones(model.rhs_eval(0, y0, inputs).shape[0])
        alg_ids = np.zeros(len(y0) - len(rhs_ids))
        ids = np.concatenate((rhs_ids, alg_ids))

        number_of_sensitivity_parameters = 0
        if model.jacp_rhs_algebraic_eval is not None:
            sensitivity_names = model.calculate_sensitivities
            if model.convert_to_format == "casadi":
                number_of_sensitivity_parameters = model.jacp_rhs_algebraic_eval.n_out()
            else:
                number_of_sensitivity_parameters = len(sensitivity_names)
        else:
            sensitivity_names = []

        # for the casadi solver we just give it dFdp_i
        if model.jacp_rhs_algebraic_eval is None:
            sensfn = casadi.Function("sensfn", [], [])
        else:
            sensfn = model.jacp_rhs_algebraic_eval

        atol = getattr(model, "atol", self.atol)
        atol = self._check_atol_type(atol, y0.size)

        # Serialize casadi functions
        idaklu_solver_fcn = idaklu.create_casadi_solver_group
        rhs_algebraic_pkl = rhs_algebraic.serialize()
        rhs_algebraic = idaklu.generate_function(rhs_algebraic_pkl)
        jac_times_cjmass_pkl = jac_times_cjmass.serialize()
        jac_times_cjmass = idaklu.generate_function(jac_times_cjmass_pkl)
        jac_rhs_algebraic_action_pkl = jac_rhs_algebraic_action.serialize()
        jac_rhs_algebraic_action = idaklu.generate_function(
            jac_rhs_algebraic_action_pkl
        )
        rootfn_pkl = rootfn.serialize()
        rootfn = idaklu.generate_function(rootfn_pkl)
        mass_action_pkl = mass_action.serialize()
        mass_action = idaklu.generate_function(mass_action_pkl)
        sensfn_pkl = sensfn.serialize()
        sensfn = idaklu.generate_function(sensfn_pkl)

        # if output_variables specified then convert 'variable' casadi
        # function expressions to idaklu-compatible functions
        self.var_idaklu_fcns = []
        self.var_idaklu_fcns_pkl = []
        self.dvar_dy_idaklu_fcns = []
        self.dvar_dy_idaklu_fcns_pkl = []
        self.dvar_dp_idaklu_fcns = []
        self.dvar_dp_idaklu_fcns_pkl = []
        for key in self.output_variables:
            self.var_idaklu_fcns_pkl.append(self.computed_var_fcns[key].serialize())
            self.var_idaklu_fcns.append(
                idaklu.generate_function(self.var_idaklu_fcns_pkl[-1])
            )
            # Convert derivative functions for sensitivities
            if (len(inputs) > 0) and (model.calculate_sensitivities):
                self.dvar_dy_idaklu_fcns_pkl.append(
                    self.computed_dvar_dy_fcns[key].serialize()
                )
                self.dvar_dy_idaklu_fcns.append(
                    idaklu.generate_function(self.dvar_dy_idaklu_fcns_pkl[-1])
                )
                self.dvar_dp_idaklu_fcns_pkl.append(
                    self.computed_dvar_dp_fcns[key].serialize()
                )
                self.dvar_dp_idaklu_fcns.append(
                    idaklu.generate_function(self.dvar_dp_idaklu_fcns_pkl[-1])
                )

        self._setup = {
            "number_of_states": len(y0),
            "inputs": len(inputs),
            "solver_function": idaklu_solver_fcn,  # callable
            "jac_bandwidth_upper": jac_bw_upper,  # int
            "jac_bandwidth_lower": jac_bw_lower,  # int
            "atol": atol,
            "rhs_algebraic": rhs_algebraic,  # function
            "rhs_algebraic_pkl": rhs_algebraic_pkl,
            "jac_times_cjmass": jac_times_cjmass,  # function
            "jac_times_cjmass_pkl": jac_times_cjmass_pkl,
            "jac_times_cjmass_colptrs": jac_times_cjmass_colptrs,  # array
            "jac_times_cjmass_rowvals": jac_times_cjmass_rowvals,  # array
            "jac_times_cjmass_nnz": jac_times_cjmass_nnz,  # int
            "jac_rhs_algebraic_action": jac_rhs_algebraic_action,  # function
            "jac_rhs_algebraic_action_pkl": jac_rhs_algebraic_action_pkl,
            "mass_action": mass_action,  # function
            "mass_action_pkl": mass_action_pkl,
            "sensfn": sensfn,  # function
            "sensfn_pkl": sensfn_pkl,
            "rootfn": rootfn,  # function
            "rootfn_pkl": rootfn_pkl,
            "num_of_events": num_of_events,  # int
            "ids": ids,  # array
            "sensitivity_names": sensitivity_names,
            "number_of_sensitivity_parameters": number_of_sensitivity_parameters,
            "standard_form_dae": model.is_standard_form_dae,  # bool
            "output_variables": self.output_variables,
            "var_fcns": self.computed_var_fcns,
            "var_idaklu_fcns": self.var_idaklu_fcns,
            "var_idaklu_fcns_pkl": self.var_idaklu_fcns_pkl,
            "dvar_dy_idaklu_fcns": self.dvar_dy_idaklu_fcns,
            "dvar_dy_idaklu_fcns_pkl": self.dvar_dy_idaklu_fcns_pkl,
            "dvar_dp_idaklu_fcns": self.dvar_dp_idaklu_fcns,
            "dvar_dp_idaklu_fcns_pkl": self.dvar_dp_idaklu_fcns_pkl,
        }

        solver = self._setup["solver_function"](
            number_of_states=self._setup["number_of_states"],
            number_of_parameters=self._setup["number_of_sensitivity_parameters"],
            rhs_alg=self._setup["rhs_algebraic"],
            jac_times_cjmass=self._setup["jac_times_cjmass"],
            jac_times_cjmass_colptrs=self._setup["jac_times_cjmass_colptrs"],
            jac_times_cjmass_rowvals=self._setup["jac_times_cjmass_rowvals"],
            jac_times_cjmass_nnz=self._setup["jac_times_cjmass_nnz"],
            jac_bandwidth_lower=self._setup["jac_bandwidth_lower"],
            jac_bandwidth_upper=self._setup["jac_bandwidth_upper"],
            jac_action=self._setup["jac_rhs_algebraic_action"],
            mass_action=self._setup["mass_action"],
            sens=self._setup["sensfn"],
            events=self._setup["rootfn"],
            number_of_events=self._setup["num_of_events"],
            rhs_alg_id=self._setup["ids"],
            atol=self._setup["atol"],
            rtol=self.rtol,
            inputs=self._setup["inputs"],
            var_fcns=self._setup["var_idaklu_fcns"],
            dvar_dy_fcns=self._setup["dvar_dy_idaklu_fcns"],
            dvar_dp_fcns=self._setup["dvar_dp_idaklu_fcns"],
            options=self._options,
        )

        self._setup["solver"] = solver

        return base_set_up_return

    def __getstate__(self):
        # if _setup is not defined then we haven't called set_up yet
        if not hasattr(self, "_setup"):
            return self.__dict__

        self.var_idaklu_fcns = []

        for key in [
            "solver",
            "solver_function",
            "rhs_algebraic",
            "jac_times_cjmass",
            "jac_rhs_algebraic_action",
            "mass_action",
            "sensfn",
            "rootfn",
            "var_idaklu_fcns",
            "dvar_dy_idaklu_fcns",
            "dvar_dp_idaklu_fcns",
        ]:
            self._setup.pop(key, None)
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

        # if _setup is not defined then we haven't called set_up yet
        if not hasattr(self, "_setup"):
            return

        self.var_idaklu_fcns = [
            idaklu.generate_function(f) for f in self.var_idaklu_fcns_pkl
        ]

        for key in [
            "rhs_algebraic",
            "jac_times_cjmass",
            "jac_rhs_algebraic_action",
            "mass_action",
            "sensfn",
            "rootfn",
        ]:
            self._setup[key] = idaklu.generate_function(self._setup[key + "_pkl"])

        for key in [
            "var_idaklu_fcns",
            "dvar_dy_idaklu_fcns",
            "dvar_dp_idaklu_fcns",
        ]:
            self._setup[key] = [
                idaklu.generate_function(f) for f in self._setup[key + "_pkl"]
            ]

        self._setup["solver_function"] = idaklu.create_casadi_solver_group

        self._setup["solver"] = self._setup["solver_function"](
            number_of_states=self._setup["number_of_states"],
            number_of_parameters=self._setup["number_of_sensitivity_parameters"],
            rhs_alg=self._setup["rhs_algebraic"],
            jac_times_cjmass=self._setup["jac_times_cjmass"],
            jac_times_cjmass_colptrs=self._setup["jac_times_cjmass_colptrs"],
            jac_times_cjmass_rowvals=self._setup["jac_times_cjmass_rowvals"],
            jac_times_cjmass_nnz=self._setup["jac_times_cjmass_nnz"],
            jac_bandwidth_lower=self._setup["jac_bandwidth_lower"],
            jac_bandwidth_upper=self._setup["jac_bandwidth_upper"],
            jac_action=self._setup["jac_rhs_algebraic_action"],
            mass_action=self._setup["mass_action"],
            sens=self._setup["sensfn"],
            events=self._setup["rootfn"],
            number_of_events=self._setup["num_of_events"],
            rhs_alg_id=self._setup["ids"],
            atol=self._setup["atol"],
            rtol=self.rtol,
            inputs=self._setup["inputs"],
            var_fcns=self._setup["var_idaklu_fcns"],
            dvar_dy_fcns=self._setup["dvar_dy_idaklu_fcns"],
            dvar_dp_fcns=self._setup["dvar_dp_idaklu_fcns"],
            options=self._options,
        )

    @property
    def supports_parallel_solve(self):
        return True

    @property
    def options(self):
        return self._options

    def _apply_solver_initial_conditions(self, model, initial_conditions):
        """
        Apply custom initial conditions to a model by overriding model.y0.

        Parameters
        ----------
        model : pybamm.BaseModel
            A model with a precomputed y0 vector.
        initial_conditions : dict or numpy.ndarray
            Either a mapping from variable names to values (scalar or array),
            or a flat numpy array matching the length of model.y0.
        """
        if isinstance(initial_conditions, dict):
            y0_np = (
                model.y0.full() if isinstance(model.y0, casadi.DM) else model.y0.copy()
            )

            for var_name, value in initial_conditions.items():
                found = False
                for symbol, slice_info in model.y_slices.items():
                    if symbol.name == var_name:
                        var_slice = slice_info[0]
                        y0_np[var_slice] = value
                        found = True
                        break
                if not found:
                    raise ValueError(f"Variable '{var_name}' not found in model")

            model.y0 = casadi.DM(y0_np)

        elif isinstance(initial_conditions, np.ndarray):
            model.y0 = casadi.DM(initial_conditions)
        else:
            raise TypeError("Initial conditions must be dict or numpy array")

    def _integrate(
        self, model, t_eval, inputs_list=None, t_interp=None, initial_conditions=None
    ):
        """
        Solve a DAE model defined by residuals with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : numeric type
            The times at which to stop the integration due to a discontinuity in time.
        inputs_list: list of dict, optional
            Any input parameters to pass to the model when solving.
        t_interp : None, list or ndarray, optional
            The times (in seconds) at which to interpolate the solution. Defaults to `None`,
            which returns the adaptive time-stepping times.
        initial_conditions : dict, numpy.ndarray, or list, optional
            Override the model’s default `y0`.  Can be:

            - a dict mapping variable names → values
            - a 1D array of length `n_states`
            - a list of such overrides (one per parallel solve)

        """
        if model.convert_to_format != "casadi":  # pragma: no cover
            # Shouldn't ever reach this point
            raise pybamm.SolverError("Unsupported IDAKLU solver configuration.")

        inputs_list = inputs_list or [{}]

        # stack inputs so that they are a 2D array of shape (number_of_inputs, number_of_parameters)
        if inputs_list and inputs_list[0]:
            inputs = np.vstack(
                [
                    np.hstack([np.array(x).reshape(-1) for x in inputs_dict.values()])
                    for inputs_dict in inputs_list
                ]
            )
        else:
            inputs = np.array([[]] * len(inputs_list))

        if initial_conditions is not None:
            if isinstance(initial_conditions, list):
                if len(initial_conditions) != len(inputs_list):
                    raise ValueError(
                        "Number of initial conditions must match number of input sets"
                    )

                y0_list = []

                model_copy = model.new_copy()
                for ic in initial_conditions:
                    self._apply_solver_initial_conditions(model_copy, ic)
                    y0_list.append(model_copy.y0.full().flatten())

                y0full = np.vstack(y0_list)
                ydot0full = np.zeros_like(y0full)

            else:
                self._apply_solver_initial_conditions(model, initial_conditions)

                y0_np = model.y0.full()

                y0full = np.vstack([y0_np for _ in range(len(inputs_list))])
                ydot0full = np.zeros_like(y0full)
        else:
            # stack y0full and ydot0full so they are a 2D array of shape (number_of_inputs, number_of_states + number_of_parameters * number_of_states)
            # note that y0full and ydot0full are currently 1D arrays (i.e. independent of inputs), but in the future we will support
            # different initial conditions for different inputs. For now we just repeat the same initial conditions for each input
            y0full = np.vstack([model.y0full] * len(inputs_list))
            ydot0full = np.vstack([model.ydot0full] * len(inputs_list))

        atol = getattr(model, "atol", self.atol)
        atol = self._check_atol_type(atol, y0full.size)

        timer = pybamm.Timer()
        solns = self._setup["solver"].solve(
            t_eval,
            t_interp,
            y0full,
            ydot0full,
            inputs,
        )
        integration_time = timer.time()

        return [
            self._post_process_solution(soln, model, integration_time, inputs_dict)
            for soln, inputs_dict in zip(solns, inputs_list, strict=False)
        ]

    def _post_process_solution(self, sol, model, integration_time, inputs_dict):
        number_of_sensitivity_parameters = self._setup[
            "number_of_sensitivity_parameters"
        ]
        sensitivity_names = self._setup["sensitivity_names"]
        number_of_timesteps = sol.t.size
        number_of_states = model.len_rhs_and_alg
        save_outputs_only = self.output_variables
        if save_outputs_only:
            # Substitute empty vectors for state vector 'y'
            y_out = np.zeros((number_of_timesteps * number_of_states, 0))
            y_event = sol.y_term
        else:
            y_out = sol.y.reshape((number_of_timesteps, number_of_states))
            y_event = y_out[-1]

        # return sensitivity solution, we need to flatten yS to
        # (#timesteps * #states (where t is changing the quickest),)
        # to match format used by Solution
        # note that yS is (n_p, n_t, n_y)
        if number_of_sensitivity_parameters != 0:
            yS_out = {
                name: sol.yS[i].reshape(-1, 1)
                for i, name in enumerate(sensitivity_names)
            }
            # add "all" stacked sensitivities ((#timesteps * #states,#sens_params))
            yS_out["all"] = np.hstack([yS_out[name] for name in sensitivity_names])
        else:
            yS_out = {}

        # 0 = solved for all t_eval
        # 2 = found root(s)
        # < 0 = solver failure
        if sol.flag == 2:
            termination = "event"
        elif sol.flag >= 0:
            termination = "final time"
        elif sol.flag < 0:
            termination = "failure"
            match self._on_failure:
                case "warn":
                    warnings.warn(
                        f"FAILURE {self._solver_flag(sol.flag)}, returning a partial solution.",
                        stacklevel=2,
                    )
                case "raise":
                    raise pybamm.SolverError(f"FAILURE {self._solver_flag(sol.flag)}")

        if sol.yp.size > 0:
            yp = sol.yp.reshape((number_of_timesteps, number_of_states)).T
        else:
            yp = None

        newsol = pybamm.Solution(
            sol.t,
            np.transpose(y_out),
            model,
            inputs_dict,
            np.array([sol.t[-1]]),
            np.transpose(y_event)[:, np.newaxis],
            termination,
            all_sensitivities=yS_out,
            all_yps=yp,
            variables_returned=bool(save_outputs_only),
        )

        newsol.integration_time = integration_time
        if not save_outputs_only:
            return newsol

        # Populate variables and sensitivities dictionaries directly
        number_of_samples = sol.y.shape[0] // number_of_timesteps
        sol.y = sol.y.reshape((number_of_timesteps, number_of_samples))
        sensitivity_params = (
            list(inputs_dict.keys()) if model.calculate_sensitivities else []
        )

        start_idx = 0
        for var in self.output_variables:
            var_nnz, var_shape, base_variables = self._get_variable_info(model, var)
            end_idx = start_idx + var_nnz
            data = sol.y[:, start_idx:end_idx]
            time_indep = False

            # handle any time integral variables
            if var in self._time_integral_vars:
                # time integral variables should all be 1D
                tiv = self._time_integral_vars[var]
                data = tiv.postfix(data.reshape(-1), sol.t, inputs_dict)
                time_indep = True

            newsol._variables[var] = pybamm.ProcessedVariableComputed(
                [model.variables_and_events[var]],
                base_variables,
                [data],
                newsol,
                time_indep=time_indep,
            )

            # Add sensitivities
            newsol[var]._sensitivities = {}
            if sensitivity_params:
                if var_nnz != math.prod(var_shape):
                    raise pybamm.SolverError(
                        f"Sensitivity of sparse variables not supported. {var} is a sparse variable with number of non-zeros {var_nnz} and shape {var_shape}"
                    )
                sens_data = sol.yS[:, start_idx:end_idx, :]
                sens_data = sens_data.reshape(
                    number_of_timesteps * (end_idx - start_idx),
                    number_of_sensitivity_parameters,
                )
                if var in self._time_integral_vars:
                    tiv = self._time_integral_vars[var]
                    sens_data = tiv.postfix_sensitivities(
                        var, data, sol.t, inputs_dict, sens_data
                    )
                newsol[var]._sensitivities["all"] = sens_data

                # Add the individual sensitivity
                for i, name in enumerate(inputs_dict.keys()):
                    sens = newsol[var]._sensitivities["all"][:, i : i + 1].reshape(-1)
                    newsol[var]._sensitivities[name] = sens

            start_idx += var_nnz
        return newsol

    def _get_variable_info(self, model, var) -> tuple:
        """Get variable length and base variables based on model format."""
        if model.convert_to_format == "casadi":
            base_var = self._setup["var_fcns"][var]
            var_eval = base_var(0.0, 0.0, 0.0)
            var_nnz = var_eval.sparsity().nnz()
            var_shape = var_eval.shape
            return var_nnz, var_shape, [base_var]
        else:  # pragma: no cover
            raise pybamm.SolverError(
                f"Unsupported evaluation engine for convert_to_format="
                f"{model.convert_to_format}"
            )

    def _set_consistent_initialization(self, model, time, inputs_dict):
        """
        Initialize y0 and ydot0 for the solver. In addition to calculating
        y0 from BaseSolver, we also calculate ydot0 for semi-explicit DAEs

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : numeric type
            The time at which to calculate the initial conditions.
        inputs_dict : dict
            Any input parameters to pass to the model when solving.
        """

        # set model.y0
        super()._set_consistent_initialization(model, time, inputs_dict)

        casadi_format = model.convert_to_format == "casadi"

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full()
        y0 = y0.flatten()

        # calculate the time derivatives of the differential equations
        # for semi-explicit DAEs
        if model.len_rhs > 0:
            ydot0 = self._rhs_dot_consistent_initialization(
                y0, model, time, inputs_dict
            )
        else:
            ydot0 = np.zeros_like(y0)

        sensitivity = (model.y0S is not None) and casadi_format
        if sensitivity:
            y0full, ydot0full = self._sensitivity_consistent_initialization(
                y0, ydot0, model, time, inputs_dict
            )
        else:
            y0full = y0
            ydot0full = ydot0

        model.y0full = y0full
        model.ydot0full = ydot0full

    def _rhs_dot_consistent_initialization(self, y0, model, time, inputs_dict):
        """
        Compute the consistent initialization of ydot0 for the differential terms
        for the solver. If we have a semi-explicit DAE, we can explicitly solve
        for this value using the consistently initialized y0 vector.

        Parameters
        ----------
        y0 : :class:`numpy.array`
            The initial values of the state vector.
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : numeric type
            The time at which to calculate the initial conditions.
        inputs_dict : dict
            Any input parameters to pass to the model when solving.

        """
        casadi_format = model.convert_to_format == "casadi"

        inputs_dict = inputs_dict or {}
        # stack inputs
        if inputs_dict:
            arrays_to_stack = [np.array(x).reshape(-1, 1) for x in inputs_dict.values()]
            inputs = np.vstack(arrays_to_stack)
        else:
            inputs = np.array([[]])

        ydot0 = np.zeros_like(y0)
        # calculate the time derivatives of the differential equations
        input_eval = inputs if casadi_format else inputs_dict

        rhs0 = model.rhs_eval(time, y0, input_eval)
        if isinstance(rhs0, casadi.DM):
            rhs0 = rhs0.full()
        rhs0 = rhs0.flatten()

        # for the differential terms, ydot = M^-1 * (rhs)
        if model.is_standard_form_dae:
            # M^-1 is the identity matrix, so we can just use rhs
            ydot0[: model.len_rhs] = rhs0
        else:
            # M^-1 is not the identity matrix, so we need to use the mass matrix
            ydot0[: model.len_rhs] = model.mass_matrix_inv.entries @ rhs0

        return ydot0

    def _sensitivity_consistent_initialization(
        self, y0, ydot0, model, time, inputs_dict
    ):
        """
        Extend the consistent initialization to include the sensitivty equations

        Parameters
        ----------
        y0 : :class:`numpy.array`
            The initial values of the state vector.
        ydot0 : :class:`numpy.array`
            The initial values of the time derivatives of the state vector.
        time : numeric type
            The time at which to calculate the initial conditions.
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        inputs_dict : dict
            Any input parameters to pass to the model when solving.

        """

        y0S = model.y0S

        if isinstance(y0S, casadi.DM):
            y0S = (y0S,)

        if isinstance(y0S[0], casadi.DM):
            y0S = (x.full() for x in y0S)
        y0S = [x.flatten() for x in y0S]

        y0full = np.concatenate([y0, *y0S])

        ydot0S = [np.zeros_like(y0S_i) for y0S_i in y0S]
        ydot0full = np.concatenate([ydot0, *ydot0S])

        return y0full, ydot0full

    def jaxify(
        self,
        model,
        t_eval,
        *,
        output_variables=None,
        calculate_sensitivities=True,
        t_interp=None,
    ):
        """JAXify the solver object

        Creates a JAX expression representing the IDAKLU-wrapped solver
        object.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model to be solved
        t_eval : numeric type, optional
            The times at which to stop the integration due to a discontinuity in time.
        output_variables : list of str, optional
            The variables to be returned. If None, all variables in the model are used.
        calculate_sensitivities : bool, optional
            Whether to calculate sensitivities. Default is True.
        t_interp : None, list or ndarray, optional
            The times (in seconds) at which to interpolate the solution. Defaults to `None`,
            which returns the adaptive time-stepping times.
        """
        obj = pybamm.IDAKLUJax(
            self,  # IDAKLU solver instance
            model,
            t_eval,
            output_variables=output_variables,
            calculate_sensitivities=calculate_sensitivities,
            t_interp=t_interp,
        )
        return obj

    @staticmethod
    def _solver_flag(flag):
        flags = {
            99: "IDA_WARNING: IDASolve succeeded but an unusual situation occurred.",
            2: "IDA_ROOT_RETURN: IDASolve succeeded and found one or more roots.",
            1: "IDA_TSTOP_RETURN: IDASolve succeeded by reaching the specified stopping point.",
            0: "IDA_SUCCESS: Successful function return.",
            -1: "IDA_TOO_MUCH_WORK: The solver took mxstep internal steps but could not reach tout.",
            -2: "IDA_TOO_MUCH_ACC: The solver could not satisfy the accuracy demanded by the user for some internal step.",
            -3: "IDA_ERR_FAIL: Error test failures occurred too many times during one internal time step or minimum step size was reached.",
            -4: "IDA_CONV_FAIL: Convergence test failures occurred too many times during one internal time step or minimum step size was reached.",
            -5: "IDA_LINIT_FAIL: The linear solver's initialization function failed.",
            -6: "IDA_LSETUP_FAIL: The linear solver's setup function failed in an unrecoverable manner.",
            -7: "IDA_LSOLVE_FAIL: The linear solver's solve function failed in an unrecoverable manner.",
            -8: "IDA_RES_FAIL: The user-provided residual function failed in an unrecoverable manner.",
            -9: "IDA_REP_RES_FAIL: The user-provided residual function repeatedly returned a recoverable error flag, but the solver was unable to recover.",
            -10: "IDA_RTFUNC_FAIL: The rootfinding function failed in an unrecoverable manner.",
            -11: "IDA_CONSTR_FAIL: The inequality constraints were violated and the solver was unable to recover.",
            -12: "IDA_FIRST_RES_FAIL: The user-provided residual function failed recoverably on the first call.",
            -13: "IDA_LINESEARCH_FAIL: The line search failed.",
            -14: "IDA_NO_RECOVERY: The residual function, linear solver setup function, or linear solver solve function had a recoverable failure, but IDACalcIC could not recover.",
            -15: "IDA_NLS_INIT_FAIL: The nonlinear solver's init routine failed.",
            -16: "IDA_NLS_SETUP_FAIL: The nonlinear solver's setup routine failed.",
            -20: "IDA_MEM_NULL: The ida mem argument was NULL.",
            -21: "IDA_MEM_FAIL: A memory allocation failed.",
            -22: "IDA_ILL_INPUT: One of the function inputs is illegal.",
            -23: "IDA_NO_MALLOC: The ida memory was not allocated by a call to IDAInit.",
            -24: "IDA_BAD_EWT: Zero value of some error weight component.",
            -25: "IDA_BAD_K: The k-th derivative is not available.",
            -26: "IDA_BAD_T: The time t is outside the last step taken.",
            -27: "IDA_BAD_DKY: The vector argument where derivative should be stored is NULL.",
        }

        flag_unknown = "Unknown IDA flag."

        return flags.get(flag, flag_unknown)



#
# Class for uniform current collectors
#
import pybamm

from .base_current_collector import BaseModel


class Uniform(BaseModel):
    """A submodel for uniform potential in the current collectors which
    is valid in the limit of fast conductivity in the current collectors.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):
        phi_s_cn = pybamm.PrimaryBroadcast(0, "current collector")
        variables = self._get_standard_negative_potential_variables(phi_s_cn)
        return variables

    def get_coupled_variables(self, variables):
        # TODO: grad not implemented for 2D yet
        i_cc = pybamm.Scalar(0)
        i_boundary_cc = pybamm.PrimaryBroadcast(
            variables["Total current density [A.m-2]"], "current collector"
        )

        variables = self._get_standard_current_variables(i_cc, i_boundary_cc)

        return variables



import os
import sys
import threading
import time
import uuid
from pathlib import Path

import platformdirs

import pybamm


def check_env_opt_out():
    return os.getenv("PYBAMM_DISABLE_TELEMETRY", "false").lower() != "false"


def check_opt_out():
    opt_out = check_env_opt_out()
    config = pybamm.config.read()
    if config:
        opt_out = opt_out or not config["enable_telemetry"]
    return opt_out


def is_running_tests():  # pragma: no cover
    """
    Detect if the code is being run as part of a test suite or building docs with Sphinx.

    Returns:
        bool: True if running tests or building docs, False otherwise.
    """
    # Check if pytest or unittest is running
    if any(
        test_module in sys.modules for test_module in ["pytest", "unittest", "nose"]
    ):
        return True

    # Check for other common CI environment variables
    ci_env_vars = [
        "GITHUB_ACTIONS",
        "CI",
        "TRAVIS",
        "CIRCLECI",
        "JENKINS_URL",
        "GITLAB_CI",
    ]
    if any(var in os.environ for var in ci_env_vars):
        return True

    # Check if building docs with Sphinx
    if any(mod == "sphinx" or mod.startswith("sphinx.") for mod in sys.modules):
        print(
            f"Found Sphinx module: {[mod for mod in sys.modules if mod.startswith('sphinx')]}"
        )
        return True

    return False


def ask_user_opt_in(timeout=10):
    """
    Ask the user if they want to opt in to telemetry.

    Parameters
    ----------
    timeout : float, optional
        The timeout for the user to respond to the prompt. Default is 10 seconds.

    Returns
    -------
    bool
        True if the user opts in, False otherwise.
    """
    print(
        "PyBaMM can collect usage data and send it to the PyBaMM team to "
        "help us improve the software.\n"
        "We do not collect any sensitive information such as models, parameters, "
        "or simulation results - only information on which parts of the code are "
        "being used and how frequently.\n"
        "This is entirely optional and does not impact the functionality of PyBaMM.\n"
        "For more information, see https://docs.pybamm.org/en/latest/source/user_guide/index.html#telemetry"
    )

    def get_input():  # pragma: no cover
        try:
            user_input = (
                input("Do you want to enable telemetry? (Y/n): ").strip().lower()
            )
            answer.append(user_input)
        except Exception:
            # Handle any input errors
            pass

    time_start = time.time()

    while True:
        if time.time() - time_start > timeout:
            print("\nTimeout reached. Defaulting to not enabling telemetry.")
            return False

        answer = []
        # Create and start input thread
        input_thread = threading.Thread(target=get_input)
        input_thread.daemon = True
        input_thread.start()

        # Wait for either timeout or input
        input_thread.join(timeout)

        if answer:
            if answer[0] in ["yes", "y", ""]:
                print("\nTelemetry enabled.\n")
                return True
            elif answer[0] in ["no", "n"]:
                print("\nTelemetry disabled.\n")
                return False
            else:
                print("\nInvalid input. Please enter 'yes/y' for yes or 'no/n' for no.")
        else:
            print("\nTimeout reached. Defaulting to not enabling telemetry.")
            return False


def generate():
    if is_running_tests() or check_opt_out():
        return

    # Check if the config file already exists
    if read() is not None:
        return

    # Ask the user if they want to opt in to telemetry
    opt_in = ask_user_opt_in()
    config_file = Path(platformdirs.user_config_dir("pybamm")) / "config.yml"
    write_uuid_to_file(config_file, opt_in)

    if opt_in:
        pybamm.telemetry.capture("user-opted-in")


def read():
    config_file = Path(platformdirs.user_config_dir("pybamm")) / "config.yml"
    return read_uuid_from_file(config_file)


def write_uuid_to_file(config_file, opt_in):
    # Create the directory if it doesn't exist
    config_file.parent.mkdir(parents=True, exist_ok=True)

    # Write the UUID to the config file in YAML format
    with open(config_file, "w") as f:
        f.write("pybamm:\n")
        f.write(f"  enable_telemetry: {opt_in}\n")
        if opt_in:
            unique_id = uuid.uuid4()
            f.write(f"  uuid: {unique_id}\n")


def read_uuid_from_file(config_file):
    # Check if the config file exists
    if not config_file.exists():
        return None

    # Read the UUID from the config file
    with open(config_file) as f:
        content = f.read().strip()

    # Extract the UUID using YAML parsing
    try:
        import yaml

        config = yaml.safe_load(content)
        return config["pybamm"]
    except (yaml.YAMLError, ValueError):
        return None



#
# Tafel classes
#
import pybamm

from .base_kinetics import BaseKinetics


class ForwardTafel(BaseKinetics):
    """
    Base submodel which implements the forward Tafel equation:

    .. math::
        j = u * j_0(c) * \\exp((ne * alpha * F * \\eta_r(c) / RT)

    Parameters
    ----------
    param :
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, reaction, options, phase="primary"):
        super().__init__(param, domain, reaction, options, phase)

    def _get_kinetics(self, j0, ne, eta_r, T, u):
        alpha = self.phase_param.alpha_bv
        Feta_RT = self.param.F * eta_r / (self.param.R * T)
        return u * j0 * pybamm.exp(ne * alpha * Feta_RT)


# backwardtafel not used by any of the models
# class BackwardTafel(BaseKinetics):
#     """
#     Base submodel which implements the backward Tafel equation:

#     .. math::
#         j = -j_0(c) * \\exp(-\\eta_r(c))

#     Parameters
#     ----------
#     param :
#         model parameters
#     domain : str
#         The domain to implement the model, either: 'Negative' or 'Positive'.


#     **Extends:** :class:`pybamm.interface.kinetics.BaseKinetics`
#     """

#     def __init__(self, param, domain, reaction, options):
#         super().__init__(param, domain, reaction, options)

#     def _get_kinetics(self, j0, ne, eta_r, T):
#         return -j0 * pybamm.exp(-(ne / (2 * (1 + self.param.Theta * T))) * eta_r)



from .steps import *
from .base_step import BaseStep, BaseStepExplicit, BaseStepImplicit
from .step_termination import BaseTermination, CurrentTermination, VoltageTermination, CustomTermination, CRateTermination, CrateTermination, _read_termination

__all__ = ['base_step', 'step_termination', 'steps']



#
# Base class for interface utilisation
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Base class for interface utilisation

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        Either 'negative' or 'positive'
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options):
        super().__init__(param, domain, options=options)

    def _get_standard_interface_utilisation_variables(self, u_var):
        domain, Domain = self.domain_Domain

        u = pybamm.maximum(u_var, 1e-8)
        u_var_av = pybamm.x_average(u_var)
        u_av = pybamm.maximum(u_var_av, 1e-8)
        if self.options.electrode_types[self.domain] == "planar":
            variables = {
                "Lithium metal interface utilisation variable": u_var_av,
                "Lithium metal interface utilisation": u_av,
            }
        else:
            variables = {
                f"{Domain} electrode interface utilisation variable": u_var,
                f"X-averaged {domain} electrode interface utilisation "
                "variable": u_var_av,
                f"{Domain} electrode interface utilisation": u,
                f"X-averaged {domain} electrode interface utilisation": u_av,
            }

        return variables



import pybamm


class VoltageModel(pybamm.BaseSubModel):
    """
    Voltage model for use with equivalent
    circuits. This model is used to calculate
    the voltage and total overpotentials
    from the other elements in the circuit.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param)
        self.model_options = options

    def get_coupled_variables(self, variables):
        ocv = variables["Open-circuit voltage [V]"]

        number_of_rc_elements = self.model_options["number of rc elements"]
        number_of_elements = number_of_rc_elements + 1

        overpotential = pybamm.Scalar(0)
        for i in range(number_of_elements):
            overpotential += variables[f"Element-{i} overpotential [V]"]

        diffusion_overpotential = variables["Diffusion overpotential [V]"]

        voltage = ocv + overpotential + diffusion_overpotential

        # Power and Resistance
        current = variables["Current [A]"]

        def x_not_zero(x):
            return ((x > 0) + (x < 0)) * x + (x >= 0) * (x <= 0)

        non_zero_current = x_not_zero(current)

        variables.update(
            {
                "Voltage [V]": voltage,
                "Overpotential [V]": overpotential,
                "Battery voltage [V]": voltage,
                "Power [W]": voltage * current,
                "Resistance [Ohm]": pybamm.sign(current) * voltage / non_zero_current,
            }
        )

        return variables

    def add_events_from(self, variables):
        voltage = variables["Voltage [V]"]

        # Add voltage events
        maximum_voltage = pybamm.Event(
            "Maximum voltage [V]",
            self.param.voltage_high_cut - voltage,
            pybamm.EventType.TERMINATION,
        )
        self.events.append(maximum_voltage)

        minimum_voltage = pybamm.Event(
            "Minimum voltage [V]",
            voltage - self.param.voltage_low_cut,
            pybamm.EventType.TERMINATION,
        )
        self.events.append(minimum_voltage)

        # Cut-off voltage for event switch with casadi 'fast with events'
        tol = 0.125
        self.events.append(
            pybamm.Event(
                "Minimum voltage switch",
                voltage - (self.param.voltage_low_cut - tol),
                pybamm.EventType.SWITCH,
            )
        )
        self.events.append(
            pybamm.Event(
                "Maximum voltage switch",
                voltage - (self.param.voltage_high_cut + tol),
                pybamm.EventType.SWITCH,
            )
        )



import json
import numbers
import re
from collections import defaultdict
from pathlib import Path
from pprint import pformat
from typing import Any
from warnings import warn

import numpy as np

import pybamm
from pybamm.expression_tree.operations.serialise import (
    Serialise,
    convert_function_to_symbolic_expression,
    convert_symbol_from_json,
    convert_symbol_to_json,
)
from pybamm.models.full_battery_models.lithium_ion.msmr import (
    is_deprecated_msmr_name,
    replace_deprecated_msmr_name,
)


class ParameterValues:
    """
    The parameter values for a simulation.

    Note that this class does not inherit directly from the python dictionary class as
    this causes issues with saving and loading simulations.

    Parameters
    ----------
    values : dict or string
        Explicit set of parameters, or reference to an inbuilt parameter set
        If string and matches one of the inbuilt parameter sets, returns that parameter
        set.

    Examples
    --------
    >>> values = {"some parameter": 1, "another parameter": 2}
    >>> param = pybamm.ParameterValues(values)
    >>> param["some parameter"]
    1
    >>> param = pybamm.ParameterValues("Marquis2019")
    >>> param["Reference temperature [K]"]
    298.15

    """

    def __init__(self, values):
        # add physical constants as default values
        self._dict_items = pybamm.FuzzyDict(
            {
                "Ideal gas constant [J.K-1.mol-1]": pybamm.constants.R.value,
                "Faraday constant [C.mol-1]": pybamm.constants.F.value,
                "Boltzmann constant [J.K-1]": pybamm.constants.k_b.value,
                "Electron charge [C]": pybamm.constants.q_e.value,
            }
        )

        if isinstance(values, dict | ParameterValues):
            # remove the "chemistry" key if it exists
            chemistry = values.pop("chemistry", None)
            self.update(values, check_already_exists=False)
        else:
            # Check if values is a named parameter set
            if isinstance(values, str) and values in pybamm.parameter_sets.keys():
                values = pybamm.parameter_sets[values]
                chemistry = values.pop("chemistry", None)
                self.update(values, check_already_exists=False)
            else:
                valid_sets = "\n".join(pybamm.parameter_sets.keys())
                raise ValueError(
                    f"'{values}' is not a valid parameter set. Parameter set must be one of:\n{valid_sets}"
                )

        if chemistry == "ecm":
            self._set_initial_state = pybamm.equivalent_circuit.set_initial_state
        else:
            self._set_initial_state = pybamm.lithium_ion.set_initial_state

        # Initialise empty _processed_symbols dict (for caching)
        self._processed_symbols = {}

        # save citations
        if "citations" in self._dict_items:
            for citation in self._dict_items["citations"]:
                pybamm.citations.register(citation)

    @staticmethod
    def _create_from_bpx(bpx, target_soc):
        from bpx import get_electrode_concentrations
        from bpx.schema import ElectrodeBlended, ElectrodeBlendedSPM

        from .bpx import bpx_to_param_dict

        if target_soc < 0 or target_soc > 1:
            raise ValueError("Target SOC should be between 0 and 1")

        pybamm_dict = bpx_to_param_dict(bpx)

        if "Open-circuit voltage at 0% SOC [V]" not in pybamm_dict:
            pybamm_dict["Open-circuit voltage at 0% SOC [V]"] = pybamm_dict[
                "Lower voltage cut-off [V]"
            ]
            warn(
                "'Open-circuit voltage at 0% SOC [V]' not found in BPX file. Using "
                "'Lower voltage cut-off [V]'.",
                stacklevel=2,
            )
        if "Open-circuit voltage at 100% SOC [V]" not in pybamm_dict:
            pybamm_dict["Open-circuit voltage at 100% SOC [V]"] = pybamm_dict[
                "Upper voltage cut-off [V]"
            ]
            warn(
                "'Open-circuit voltage at 100% SOC [V]' not found in BPX file. Using "
                "'Upper voltage cut-off [V]'.",
                stacklevel=2,
            )

        # get initial concentrations based on SOC
        # Note: we cannot set SOC for blended electrodes,
        # see https://github.com/pybamm-team/PyBaMM/issues/2682
        bpx_neg = bpx.parameterisation.negative_electrode
        bpx_pos = bpx.parameterisation.positive_electrode
        if isinstance(bpx_neg, ElectrodeBlended | ElectrodeBlendedSPM) or isinstance(
            bpx_pos, ElectrodeBlended | ElectrodeBlendedSPM
        ):
            pybamm.logger.warning(
                "Initial concentrations cannot be set using stoichiometry limits for "
                "blend electrodes. Please set the initial concentrations manually."
            )
        else:
            c_n_init, c_p_init = get_electrode_concentrations(target_soc, bpx)
            pybamm_dict["Initial concentration in negative electrode [mol.m-3]"] = (
                c_n_init
            )
            pybamm_dict["Initial concentration in positive electrode [mol.m-3]"] = (
                c_p_init
            )

        return pybamm.ParameterValues(pybamm_dict)

    @staticmethod
    def create_from_bpx_obj(bpx_obj, target_soc: float = 1):
        """
        Parameters
        ----------
        bpx_obj: dict
            A dictionary containing the parameters in the `BPX <https://bpxstandard.com/>`_ format
        target_soc : float, optional
            Target state of charge. Must be between 0 and 1. Default is 1.

        Returns
        -------
        ParameterValues
            A parameter values object with the parameters in the bpx file

        """
        from bpx import parse_bpx_obj

        bpx = parse_bpx_obj(bpx_obj)
        return ParameterValues._create_from_bpx(bpx, target_soc)

    @staticmethod
    def create_from_bpx(filename, target_soc: float = 1):
        """
        Parameters
        ----------
        filename: str
            The filename of the `BPX <https://bpxstandard.com/>`_ file
        target_soc : float, optional
            Target state of charge. Must be between 0 and 1. Default is 1.

        Returns
        -------
        ParameterValues
            A parameter values object with the parameters in the bpx file

        """
        from bpx import parse_bpx_file

        bpx = parse_bpx_file(filename)
        return ParameterValues._create_from_bpx(bpx, target_soc)

    def __getitem__(self, key):
        try:
            return self._dict_items[key]
        except KeyError as err:
            if (
                "Exchange-current density for lithium metal electrode [A.m-2]"
                in err.args[0]
                and "Exchange-current density for plating [A.m-2]" in self._dict_items
            ):
                raise KeyError(
                    "'Exchange-current density for plating [A.m-2]' has been renamed "
                    "to 'Exchange-current density for lithium metal electrode [A.m-2]' "
                    "when referring to the reaction at the surface of a lithium metal "
                    "electrode. This is to avoid confusion with the exchange-current "
                    "density for the lithium plating reaction in a porous negative "
                    "electrode. To avoid this error, change your parameter file to use "
                    "the new name."
                ) from err
            else:
                raise err

    def get(self, key, default=None):
        """Return item corresponding to key if it exists, otherwise return default"""
        try:
            return self._dict_items[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        """Call the update functionality when doing a setitem"""
        self.update({key: value})

    def __delitem__(self, key):
        del self._dict_items[key]

    def __repr__(self):
        return pformat(self._dict_items, width=1)

    def __eq__(self, other):
        return self._dict_items == other._dict_items

    def keys(self):
        """Get the keys of the dictionary"""
        return self._dict_items.keys()

    def values(self):
        """Get the values of the dictionary"""
        return self._dict_items.values()

    def items(self):
        """Get the items of the dictionary"""
        return self._dict_items.items()

    def pop(self, *args, **kwargs):
        return self._dict_items.pop(*args, **kwargs)

    def copy(self):
        """Returns a copy of the parameter values. Makes sure to copy the internal
        dictionary."""
        new_copy = ParameterValues(self._dict_items.copy())
        new_copy._set_initial_state = self._set_initial_state
        return new_copy

    def search(self, key, print_values=True):
        """
        Search dictionary for keys containing 'key'.

        See :meth:`pybamm.FuzzyDict.search()`.
        """
        return self._dict_items.search(key, print_values)

    def update(self, values, check_conflict=False, check_already_exists=True, path=""):
        """
        Update parameter dictionary, while also performing some basic checks.

        Parameters
        ----------
        values : dict
            Dictionary of parameter values to update parameter dictionary with
        check_conflict : bool, optional
            Whether to check that a parameter in `values` has not already been defined
            in the parameter class when updating it, and if so that its value does not
            change. This is set to True during initialisation, when parameters are
            combined from different sources, and is False by default otherwise
        check_already_exists : bool, optional
            Whether to check that a parameter in `values` already exists when trying to
            update it. This is to avoid cases where an intended change in the parameters
            is ignored due a typo in the parameter name, and is True by default but can
            be manually overridden.
        path : string, optional
            Path from which to load functions
        """
        # check if values is not a dictionary
        if not isinstance(values, dict):
            values = values._dict_items
        # check parameter values
        values = self.check_parameter_values(values)
        # update
        for name, value in values.items():
            # check for conflicts
            if (
                check_conflict is True
                and name in self.keys()
                and not (self[name] == float(value) or self[name] == value)
            ):
                raise ValueError(
                    f"parameter '{name}' already defined with value '{self[name]}'"
                )
            # check parameter already exists (for updating parameters)
            if check_already_exists is True:
                try:
                    self._dict_items[name]
                except KeyError as err:
                    raise KeyError(
                        f"Cannot update parameter '{name}' as it does not "
                        + f"have a default value. ({err.args[0]}). If you are "
                        + "sure you want to update this parameter, use "
                        + "param.update({name: value}, check_already_exists=False)"
                    ) from err
            if isinstance(value, str):
                if (
                    value.startswith("[function]")
                    or value.startswith("[current data]")
                    or value.startswith("[data]")
                    or value.startswith("[2D data]")
                ):
                    raise ValueError(
                        "Specifying parameters via [function], [current data], [data] "
                        "or [2D data] is no longer supported. For functions, pass in a "
                        "python function object. For data, pass in a python function "
                        "that returns a pybamm Interpolant object. "
                        "See the Ai2020 parameter set for an example with both."
                    )

                elif value == "[input]":
                    self._dict_items[name] = pybamm.InputParameter(name)
                # Anything else should be a converted to a float
                else:
                    self._dict_items[name] = float(value)
            elif isinstance(value, tuple) and isinstance(value[1], np.ndarray):
                # If data is provided as a 2-column array (1D data),
                # convert to two arrays for compatibility with 2D data
                # see #1805
                func_name, data = value
                data = ([data[:, 0]], data[:, 1])
                self._dict_items[name] = (func_name, data)
            else:
                self._dict_items[name] = value
        # reset processed symbols
        self._processed_symbols = {}

    def set_initial_state(
        self,
        initial_value,
        direction=None,
        param=None,
        inplace=True,
        options=None,
        inputs=None,
    ):
        return self._set_initial_state(
            initial_value,
            self,
            direction=direction,
            param=param,
            inplace=inplace,
            options=options,
            inputs=inputs,
        )

    def set_initial_stoichiometry_half_cell(
        self,
        initial_value,
        direction=None,
        param=None,
        known_value="cyclable lithium capacity",
        inplace=True,
        options=None,
        inputs=None,
    ):
        msg = "pybamm.parameter_values.set_initial_stoichiometry_half_cell is deprecated, please use set_initial_state."
        warn(msg, DeprecationWarning, stacklevel=2)
        return self._set_initial_state(
            initial_value,
            self,
            direction=direction,
            param=param,
            known_value=known_value,
            inplace=inplace,
            options=options,
            inputs=inputs,
        )

    def set_initial_stoichiometries(
        self,
        initial_value,
        direction=None,
        param=None,
        known_value="cyclable lithium capacity",
        inplace=True,
        options=None,
        inputs=None,
        tol=1e-6,
    ):
        msg = "pybamm.parameter_values.set_initial_stoichiometries is deprecated, please use set_initial_state."
        warn(msg, DeprecationWarning, stacklevel=2)
        return self._set_initial_state(
            initial_value,
            self,
            direction=direction,
            param=param,
            known_value=known_value,
            inplace=inplace,
            options=options,
            inputs=inputs,
            tol=tol,
        )

    def set_initial_ocps(
        self,
        initial_value,
        direction=None,
        param=None,
        known_value="cyclable lithium capacity",
        inplace=True,
        options=None,
        inputs=None,
    ):
        msg = "pybamm.parameter_values.set_initial_ocps is deprecated, please use set_initial_state."
        warn(msg, DeprecationWarning, stacklevel=2)
        return self._set_initial_state(
            initial_value,
            self,
            direction=direction,
            param=param,
            known_value=known_value,
            inplace=inplace,
            options=options,
            inputs=inputs,
        )

    @staticmethod
    def check_parameter_values(values):
        values = scalarize_dict(values)
        for param in list(values.keys()):
            if "propotional term" in param:
                raise ValueError(
                    f"The parameter '{param}' has been renamed to "
                    "'... proportional term [s-1]', and its value should now be divided"
                    "by 3600 to get the same results as before."
                )
            # specific check for renamed parameter "1 + dlnf/dlnc"
            if "1 + dlnf/dlnc" in param:
                raise ValueError(
                    f"parameter '{param}' has been renamed to 'Thermodynamic factor'"
                )
            if "electrode diffusivity" in param:
                new_param = param.replace("electrode", "particle")
                warn(
                    f"The parameter '{param}' has been renamed to '{new_param}'",
                    DeprecationWarning,
                    stacklevel=2,
                )
                values[new_param] = values.get(param)
            if is_deprecated_msmr_name(param):
                new_param = replace_deprecated_msmr_name(param)
                warn(
                    f"The parameter '{param}' has been renamed to '{new_param}'",
                    DeprecationWarning,
                    stacklevel=2,
                )
                values[new_param] = values.get(param)

        return values

    def process_model(self, unprocessed_model, inplace=True):
        """Assign parameter values to a model.
        Currently inplace, could be changed to return a new model.

        Parameters
        ----------
        unprocessed_model : :class:`pybamm.BaseModel`
            Model to assign parameter values for
        inplace: bool, optional
            If True, replace the parameters in the model in place. Otherwise, return a
            new model with parameter values set. Default is True.

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic = {}` and
            `model.variables = {}`)

        """
        pybamm.logger.info(f"Start setting parameters for {unprocessed_model.name}")

        # set up inplace vs not inplace
        if inplace:
            # any changes to unprocessed_model attributes will change model attributes
            # since they point to the same object
            model = unprocessed_model
        else:
            # create a copy of the model
            model = unprocessed_model.new_copy()

        if (
            len(unprocessed_model.rhs) == 0
            and len(unprocessed_model.algebraic) == 0
            and len(unprocessed_model.variables) == 0
        ):
            raise pybamm.ModelError("Cannot process parameters for empty model")

        new_rhs = {}
        for variable, equation in unprocessed_model.rhs.items():
            pybamm.logger.verbose(f"Processing parameters for {variable!r} (rhs)")
            new_variable = self.process_symbol(variable)
            new_rhs[new_variable] = self.process_symbol(equation)
        model.rhs = new_rhs

        new_algebraic = {}
        for variable, equation in unprocessed_model.algebraic.items():
            pybamm.logger.verbose(f"Processing parameters for {variable!r} (algebraic)")
            new_variable = self.process_symbol(variable)
            new_algebraic[new_variable] = self.process_symbol(equation)
        model.algebraic = new_algebraic

        new_initial_conditions = {}
        for variable, equation in unprocessed_model.initial_conditions.items():
            pybamm.logger.verbose(
                f"Processing parameters for {variable!r} (initial conditions)"
            )
            new_variable = self.process_symbol(variable)
            new_initial_conditions[new_variable] = self.process_symbol(equation)
        model.initial_conditions = new_initial_conditions

        model.boundary_conditions = self.process_boundary_conditions(unprocessed_model)

        new_variables = {}
        for variable, equation in unprocessed_model.variables.items():
            pybamm.logger.verbose(f"Processing parameters for {variable!r} (variables)")
            new_variables[variable] = self.process_symbol(equation)
        model.variables = new_variables

        new_events = []
        for event in unprocessed_model.events:
            pybamm.logger.verbose(f"Processing parameters for event '{event.name}''")
            new_events.append(
                pybamm.Event(
                    event.name, self.process_symbol(event.expression), event.event_type
                )
            )

        interpolant_events = self._get_interpolant_events(model)
        for event in interpolant_events:
            pybamm.logger.verbose(f"Processing parameters for event '{event.name}''")
            new_events.append(
                pybamm.Event(
                    event.name, self.process_symbol(event.expression), event.event_type
                )
            )

        model.events = new_events

        pybamm.logger.info(f"Finish setting parameters for {model.name}")

        return model

    def _get_interpolant_events(self, model):
        """Add events for functions that have been defined as parameters"""
        # Define events to catch extrapolation. In these events the sign is
        # important: it should be positive inside of the range and negative
        # outside of it
        interpolants = model._find_symbols(pybamm.Interpolant)
        interpolant_events = []
        for interpolant in interpolants:
            xs = interpolant.x
            children = interpolant.children
            for x, child in zip(xs, children, strict=False):
                interpolant_events.extend(
                    [
                        pybamm.Event(
                            f"Interpolant '{interpolant.name}' lower bound",
                            pybamm.min(child - min(x)),
                            pybamm.EventType.INTERPOLANT_EXTRAPOLATION,
                        ),
                        pybamm.Event(
                            f"Interpolant '{interpolant.name}' upper bound",
                            pybamm.min(max(x) - child),
                            pybamm.EventType.INTERPOLANT_EXTRAPOLATION,
                        ),
                    ]
                )
        return interpolant_events

    def process_boundary_conditions(self, model):
        """
        Process boundary conditions for a model
        Boundary conditions are dictionaries {"left": left bc, "right": right bc}
        in general, but may be imposed on the tabs (or *not* on the tab) for a
        small number of variables, e.g. {"negative tab": neg. tab bc,
        "positive tab": pos. tab bc "no tab": no tab bc}.
        """
        new_boundary_conditions = {}
        sides = [
            "left",
            "right",
            "negative tab",
            "positive tab",
            "no tab",
            "top",
            "bottom",
            "x_min",
            "x_max",
            "y_min",
            "y_max",
            "z_min",
            "z_max",
            "r_min",
            "r_max",
        ]
        for variable, bcs in model.boundary_conditions.items():
            processed_variable = self.process_symbol(variable)
            new_boundary_conditions[processed_variable] = {}
            for side in sides:
                try:
                    bc, typ = bcs[side]
                    pybamm.logger.verbose(
                        f"Processing parameters for {variable!r} ({side} bc)"
                    )
                    processed_bc = (self.process_symbol(bc), typ)
                    new_boundary_conditions[processed_variable][side] = processed_bc
                except KeyError as err:
                    # don't raise error if the key error comes from the side not being
                    # found
                    if err.args[0] in side:
                        pass
                    # do raise error otherwise (e.g. can't process symbol)
                    else:
                        raise err

        return new_boundary_conditions

    def process_geometry(self, geometry):
        """
        Assign parameter values to a geometry (inplace).

        Parameters
        ----------
        geometry : dict
            Geometry specs to assign parameter values to
        """

        def process_and_check(sym):
            new_sym = self.process_symbol(sym)
            leaves = new_sym.post_order(filter=lambda node: len(node.children) == 0)
            for leaf in leaves:
                if not isinstance(leaf, pybamm.Scalar) and not isinstance(
                    leaf, pybamm.InputParameter
                ):
                    raise ValueError(
                        "Geometry parameters must be Scalars or InputParameters after parameter processing"
                    )
            return new_sym

        for domain in geometry:
            for spatial_variable, spatial_limits in geometry[domain].items():
                # process tab information if using 1 or 2D current collectors
                if spatial_variable == "tabs":
                    for tab, position_info in spatial_limits.items():
                        for position_size, sym in position_info.items():
                            geometry[domain]["tabs"][tab][position_size] = (
                                process_and_check(sym)
                            )
                else:
                    for lim, sym in spatial_limits.items():
                        geometry[domain][spatial_variable][lim] = process_and_check(sym)

    def process_symbol(self, symbol):
        """Walk through the symbol and replace any Parameter with a Value.
        If a symbol has already been processed, the stored value is returned.

        Parameters
        ----------
        symbol : :class:`pybamm.Symbol`
            Symbol or Expression tree to set parameters for

        Returns
        -------
        symbol : :class:`pybamm.Symbol`
            Symbol with Parameter instances replaced by Value

        """
        try:
            return self._processed_symbols[symbol]
        except KeyError:
            if not isinstance(symbol, pybamm.FunctionParameter):
                processed_symbol = self._process_symbol(symbol)
            else:
                processed_symbol = self._process_function_parameter(symbol)
            self._processed_symbols[symbol] = processed_symbol

            return processed_symbol

    def _process_symbol(self, symbol):
        """See :meth:`ParameterValues.process_symbol()`."""

        if isinstance(symbol, pybamm.Parameter):
            value = self[symbol.name]
            if isinstance(value, numbers.Number):
                # Check not NaN (parameter in csv file but no value given)
                if np.isnan(value):
                    raise ValueError(f"Parameter '{symbol.name}' not found")
                # Scalar inherits name
                return pybamm.Scalar(value, name=symbol.name)
            elif isinstance(value, pybamm.Symbol):
                new_value = self.process_symbol(value)
                new_value.copy_domains(symbol)
                return new_value
            else:
                raise TypeError(f"Cannot process parameter '{value}'")

        elif isinstance(symbol, pybamm.FunctionParameter):
            function_name = self[symbol.name]
            if isinstance(
                function_name,
                numbers.Number | pybamm.Interpolant | pybamm.InputParameter,
            ) or (
                isinstance(function_name, pybamm.Symbol)
                and function_name.size_for_testing == 1
            ):
                # no need to process children, they will only be used for shape
                new_children = symbol.children
            else:
                # process children
                new_children = []
                for child in symbol.children:
                    if symbol.diff_variable is not None and any(
                        x == symbol.diff_variable for x in child.pre_order()
                    ):
                        # Wrap with NotConstant to avoid simplification,
                        # which would stop symbolic diff from working properly
                        new_child = pybamm.NotConstant(child)
                        new_children.append(self.process_symbol(new_child))
                    else:
                        new_children.append(self.process_symbol(child))

            # Create Function or Interpolant or Scalar object
            if isinstance(function_name, tuple):
                if len(function_name) == 2:  # CSV or JSON parsed data
                    # to create an Interpolant
                    name, data = function_name

                    if len(data[0]) == 1:
                        input_data = data[0][0], data[1]

                    else:
                        input_data = data

                    # For parameters provided as data we use a cubic interpolant
                    # Note: the cubic interpolant can be differentiated
                    function = pybamm.Interpolant(
                        input_data[0],
                        input_data[-1],
                        new_children,
                        name=name,
                    )

                else:  # pragma: no cover
                    raise ValueError(
                        f"Invalid function name length: {len(function_name)}"
                    )

            elif isinstance(function_name, numbers.Number):
                # Check not NaN (parameter in csv file but no value given)
                if np.isnan(function_name):
                    raise ValueError(
                        f"Parameter '{symbol.name}' (possibly a function) not found"
                    )
                # If the "function" is provided is actually a scalar, return a Scalar
                # object instead of throwing an error.
                function = pybamm.Scalar(function_name, name=symbol.name)
            elif callable(function_name):
                # otherwise evaluate the function to create a new PyBaMM object
                function = function_name(*new_children)
            elif isinstance(
                function_name, pybamm.Interpolant | pybamm.InputParameter
            ) or (
                isinstance(function_name, pybamm.Symbol)
                and function_name.size_for_testing == 1
            ):
                function = function_name
            else:
                raise TypeError(
                    f"Parameter provided for '{symbol.name}' "
                    + "is of the wrong type (should either be scalar-like or callable)"
                )
            # Differentiate if necessary
            if symbol.diff_variable is None:
                # Use ones_like so that we get the right shapes
                function_out = function * pybamm.ones_like(*new_children)
            else:
                # return differentiated function
                new_diff_variable = self.process_symbol(symbol.diff_variable)
                function_out = function.diff(new_diff_variable)
            # Process again just to be sure
            return self.process_symbol(function_out)

        # Unary operators
        elif isinstance(symbol, pybamm.UnaryOperator):
            new_child = self.process_symbol(symbol.child)
            new_symbol = symbol.create_copy(new_children=[new_child])
            # x_average can sometimes create a new symbol with electrode thickness
            # parameters, so we process again to make sure these parameters are set
            if isinstance(symbol, pybamm.XAverage) and not isinstance(
                new_symbol, pybamm.XAverage
            ):
                new_symbol = self.process_symbol(new_symbol)
            # f_a_dist in the size average needs to be processed
            if isinstance(new_symbol, pybamm.SizeAverage):
                new_symbol.f_a_dist = self.process_symbol(new_symbol.f_a_dist)
            # position in evaluate at needs to be processed, and should be a Scalar
            if isinstance(new_symbol, pybamm.EvaluateAt):
                new_symbol_position = self.process_symbol(new_symbol.position)
                if not isinstance(new_symbol_position, pybamm.Scalar):
                    raise ValueError(
                        "'position' in 'EvaluateAt' must evaluate to a scalar"
                    )
                else:
                    new_symbol.position = new_symbol_position
            return new_symbol

        # Functions, BinaryOperators & Concatenations
        elif (
            isinstance(symbol, pybamm.Function)
            or isinstance(symbol, pybamm.Concatenation)
            or isinstance(symbol, pybamm.BinaryOperator)
        ):
            new_children = [self.process_symbol(child) for child in symbol.children]
            return symbol.create_copy(new_children)

        elif isinstance(symbol, pybamm.VectorField):
            left_symbol = self.process_symbol(symbol.lr_field)
            right_symbol = self.process_symbol(symbol.tb_field)
            return symbol.create_copy(new_children=[left_symbol, right_symbol])

        # Variables: update scale
        elif isinstance(symbol, pybamm.Variable):
            new_symbol = symbol.create_copy()
            new_symbol._scale = self.process_symbol(symbol.scale)
            reference = self.process_symbol(symbol.reference)
            if isinstance(reference, pybamm.Vector):
                # address numpy 1.25 deprecation warning: array should have ndim=0
                # before conversion
                reference = pybamm.Scalar((reference.evaluate()).item())
            new_symbol._reference = reference
            new_symbol.bounds = tuple([self.process_symbol(b) for b in symbol.bounds])
            return new_symbol

        elif isinstance(symbol, numbers.Number):
            return pybamm.Scalar(symbol)

        else:
            # Backup option: return the object
            return symbol

    def _process_function_parameter(self, symbol):
        function_parameter = self[symbol.name]
        # Handle symbolic function parameter case
        if isinstance(function_parameter, pybamm.ExpressionFunctionParameter):
            # Process children
            new_children = []
            for child in symbol.children:
                if symbol.diff_variable is not None and any(
                    x == symbol.diff_variable for x in child.pre_order()
                ):
                    # Wrap with NotConstant to avoid simplification,
                    # which would stop symbolic diff from working properly
                    new_child = pybamm.NotConstant(child)
                    new_children.append(self.process_symbol(new_child))
                else:
                    new_children.append(self.process_symbol(child))

            # Get the expression and inputs for the function
            expression = function_parameter.child
            inputs = {
                arg: child
                for arg, child in zip(
                    function_parameter.func_args, symbol.children, strict=True
                )
            }

            # Set domains for function inputs in post-order traversal
            for node in expression.post_order():
                if node.name in inputs:
                    node.domains = inputs[node.name].domains
                else:
                    node.domains = node.get_children_domains(node.children)

            # Combine parameter values with inputs
            combined_params = ParameterValues({**self, **inputs})

            # Process any FunctionParameter children first to avoid recursion
            for child in expression.pre_order():
                if isinstance(child, pybamm.FunctionParameter):
                    # Build new child with parent inputs
                    new_child_children = [
                        inputs[child_child.name]
                        if isinstance(child_child, pybamm.Parameter)
                        and child_child.name in inputs
                        else child_child
                        for child_child in child.children
                    ]
                    new_child = pybamm.FunctionParameter(
                        child.name,
                        dict(zip(child.input_names, new_child_children, strict=False)),
                        diff_variable=child.diff_variable,
                        print_name=child.print_name,
                    )

                    # For this local combined parameter values, process the new child
                    # and store the result as the processed symbol for this child
                    # This means the child is evaluated with the parent inputs only when
                    # it is called from within the parent function (not elsewhere in
                    # the expression tree)
                    combined_params._processed_symbols[child] = (
                        combined_params.process_symbol(new_child)
                    )

            # Process function with combined parameter values to get a symbolic expression
            function = combined_params.process_symbol(expression)

            # Differentiate if necessary
            if symbol.diff_variable is None:
                # Use ones_like so that we get the right shapes
                function_out = function * pybamm.ones_like(*new_children)
            else:
                # return differentiated function
                new_diff_variable = self.process_symbol(symbol.diff_variable)
                function_out = function.diff(new_diff_variable)

            return function_out

        # Handle non-symbolic function_name case
        else:
            return self._process_symbol(symbol)

    def evaluate(self, symbol, inputs=None):
        """
        Process and evaluate a symbol.

        Parameters
        ----------
        symbol : :class:`pybamm.Symbol`
            Symbol or Expression tree to evaluate

        Returns
        -------
        number or array
            The evaluated symbol
        """
        processed_symbol = self.process_symbol(symbol)
        if processed_symbol.is_constant():
            return processed_symbol.evaluate()
        else:
            # In the case that the only issue is an input parameter contained in inputs,
            # go ahead and try and evaluate it with the inputs. If it doesn't work, raise
            # the value error.
            try:
                return processed_symbol.evaluate(inputs=inputs)
            except Exception as exc:
                raise ValueError(
                    "symbol must evaluate to a constant scalar or array"
                ) from exc

    def _ipython_key_completions_(self):
        return list(self._dict_items.keys())

    def print_parameters(self, parameters, output_file=None):
        """
        Return dictionary of evaluated parameters, and optionally print these evaluated
        parameters to an output file.

        Parameters
        ----------
        parameters : class or dict containing :class:`pybamm.Parameter` objects
            Class or dictionary containing all the parameters to be evaluated
        output_file : string, optional
            The file to print parameters to. If None, the parameters are not printed,
            and this function simply acts as a test that all the parameters can be
            evaluated, and returns the dictionary of evaluated parameters.

        Returns
        -------
        evaluated_parameters : defaultdict
            The evaluated parameters, for further processing if needed

        Notes
        -----
        A C-rate of 1 C is the current required to fully discharge the battery in 1
        hour, 2 C is current to discharge the battery in 0.5 hours, etc
        """
        # Set list of attributes to ignore, for when we are evaluating parameters from
        # a class of parameters
        ignore = [
            "__name__",
            "__doc__",
            "__package__",
            "__loader__",
            "__spec__",
            "__file__",
            "__cached__",
            "__builtins__",
            "absolute_import",
            "division",
            "print_function",
            "unicode_literals",
            "pybamm",
            "_options",
            "constants",
            "np",
            "geo",
            "elec",
            "therm",
            "half_cell",
            "x",
            "r",
        ]

        # If 'parameters' is a class, extract the dict
        if not isinstance(parameters, dict):
            parameters_dict = {
                k: v for k, v in parameters.__dict__.items() if k not in ignore
            }
            for domain in ["n", "s", "p"]:
                domain_param = getattr(parameters, domain)
                parameters_dict.update(
                    {
                        f"{domain}.{k}": v
                        for k, v in domain_param.__dict__.items()
                        if k not in ignore
                    }
                )
            parameters = parameters_dict

        evaluated_parameters = defaultdict(list)

        # Turn to regular dictionary for faster KeyErrors
        self._dict_items = dict(self._dict_items)

        for name, symbol in parameters.items():
            if isinstance(symbol, pybamm.Symbol):
                try:
                    proc_symbol = self.process_symbol(symbol)
                except KeyError:
                    # skip parameters that don't have a value in that parameter set
                    proc_symbol = None
                if not (
                    callable(proc_symbol)
                    or proc_symbol is None
                    or proc_symbol.has_symbol_of_classes(
                        (pybamm.Concatenation, pybamm.Broadcast)
                    )
                ):
                    evaluated_parameters[name] = proc_symbol.evaluate(t=0)

            # Turn back to FuzzyDict
            self._dict_items = pybamm.FuzzyDict(self._dict_items)

        # Print the evaluated_parameters dict to output_file
        if output_file:
            self.print_evaluated_parameters(evaluated_parameters, output_file)

        return evaluated_parameters

    def print_evaluated_parameters(self, evaluated_parameters, output_file):
        """
        Print a dictionary of evaluated parameters to an output file

        Parameters
        ----------
        evaluated_parameters : defaultdict
            The evaluated parameters, for further processing if needed
        output_file : string, optional
            The file to print parameters to. If None, the parameters are not printed,
            and this function simply acts as a test that all the parameters can be
            evaluated

        """
        # Get column width for pretty printing
        column_width = max(len(name) for name in evaluated_parameters.keys())
        s = f"{{:>{column_width}}}"
        with open(output_file, "w") as file:
            for name, value in sorted(evaluated_parameters.items()):
                if 0.001 < abs(value) < 1000:
                    file.write((s + " : {:10.4g}\n").format(name, value))
                else:
                    file.write((s + " : {:10.3E}\n").format(name, value))

    def __contains__(self, key):
        return key in self._dict_items

    def __iter__(self):
        return iter(self._dict_items)

    @staticmethod
    def from_json(filename_or_dict):
        """
        Loads a ParameterValues object from a JSON file or a dictionary.

        Parameters
        ----------
        filename_or_dict : string-like or dict
            The filename to load the JSON file from, or a dictionary.

        Returns
        -------
        ParameterValues
            The ParameterValues object
        """
        if isinstance(filename_or_dict, str | Path):
            with open(filename_or_dict) as f:
                parameter_values_dict = json.load(f)
        elif isinstance(filename_or_dict, dict):
            parameter_values_dict = filename_or_dict.copy()
        else:
            raise TypeError("Input must be a filename (str or pathlib.Path) or a dict")

        for key, value in parameter_values_dict.items():
            if isinstance(value, dict):
                parameter_values_dict[key] = convert_symbol_from_json(value)

        return ParameterValues(parameter_values_dict)

    def to_json(self, filename=None):
        """
        Converts the parameter values to a JSON-serializable dictionary and optionally
        saves it to a file.

        Parameters
        ----------
        filename : str, optional
            The filename to save the JSON file to. If not provided, the dictionary is
            not saved.

        Returns
        -------
        dict
            The JSON-serializable dictionary
        """
        return convert_parameter_values_to_json(self, filename)


_SCALAR_KEY_RE = re.compile(
    r""" ^
         (?P<base>[^\[\]]+?)           # foo
         \s\((?P<idx>\d+)\)            # (0)
         (?: \s\[(?P<tag>[^\]]+)\])?   # optional [unit]
         $ """,
    re.VERBOSE,
)


def convert_symbols_in_dict(
    data_dict: dict | None = None,
) -> dict:
    """Recursively converts nested dicts using convert_symbol_from_json."""
    # Create parameter values
    if data_dict:
        for key, value in data_dict.items():
            if isinstance(value, dict) and "interpolator" in value:
                # handle interpolant
                interpolator = value.get("interpolator", "linear")
                x = value.get("x", [])
                y = value.get("y", [])

                # Convert list to pybamm.Interpolant
                def interpolant_function(sto, x=x, y=y, interpolator=interpolator):
                    try:
                        return pybamm.Interpolant(x, y, sto, interpolator=interpolator)
                    except Exception as e:
                        print(e)
                        return pybamm.Scalar(0)

                data_dict[key] = interpolant_function
            elif isinstance(value, dict):
                # Handle function parameters in JSON format
                # Recursively process nested dictionaries
                data_dict[key] = convert_symbol_from_json(value)
            elif isinstance(value, list):
                data_dict[key] = [
                    convert_symbol_from_json(item) if isinstance(item, dict) else item
                    for item in value
                ]
            elif isinstance(value, str):
                # Handle function parameters in string format
                data_dict[key] = float(value)
            # Keep other types as is
    else:
        # Return an empty dict if input is None
        data_dict = {}

    return data_dict


class _KeyMatch:
    """
    Match a parameter name against the key grammar.
    """

    __slots__ = ["base", "idx", "is_match", "name", "tag"]

    def __init__(self, name: str):
        if not isinstance(name, str):
            raise ValueError("name must be a string")
        self.name = name

        match = _SCALAR_KEY_RE.match(name)
        self.is_match = bool(match)

        if match:
            base = match["base"]
            idx = int(match["idx"])
            tag = match["tag"]
        else:
            base = ""
            idx = -1
            tag = ""

        self.base = base
        self.idx = idx
        self.tag = tag

    def __bool__(self):
        return self.is_match


def scalarize_dict(
    params: dict[str, Any], ignored_keys: list[str] | None = None
) -> dict[str, Any]:
    """
    Expand list-valued items into scalar keys while preserving tags.
    Example
    -------
    {'a [V]': [1, 2]}  →  {'a (0) [V]': 1, 'a (1) [V]': 2}

    Parameters
    ----------
    params : dict[str, Any]
        The dictionary to scalarize
    ignored_keys : list[str], optional
        The keys to ignore. Defaults to ``["citations"]``.

    Returns
    -------
    dict[str, Any]
        The scalarized dictionary
    """
    out = {}

    # Default ignored keys. Special case for citations in `pybamm.ParameterValues`
    if ignored_keys is None:
        ignored_keys = ["citations"]

    for key, val in params.items():
        if key not in ignored_keys and isinstance(val, list):
            base, tag = _split_key(key)  # accepts 'a' or 'a [V]'
            for i, item in enumerate(val):
                key_i = _combine_name(base, i, tag)
                if key_i in out:
                    raise ValueError(f"Duplicate key {key_i!r}")
                out[key_i] = item
        else:
            if key in out:
                raise ValueError(f"Duplicate key {key!r}")
            out[key] = val
    return out


def _is_iterable(val: Any) -> bool:
    return hasattr(val, "__iter__") and not isinstance(val, (str | dict | bytes))


_KEY_RE = re.compile(
    r""" ^
         (?P<base>[^\[\]]+?)           # foo
         (?: \s\[(?P<tag>[^\]]+)\])?   # optional [unit]   (for collapsed keys)
         $ """,
    re.VERBOSE,
)


def _split_key(name: str) -> tuple[str, str | None]:
    """
    Separate *name* into ``(base, tag)`` where *tag* can be ``None``.
    Works for either collapsed or scalar keys.
    """
    m = _KEY_RE.match(name)
    if not m:
        raise ValueError(f"Illegal parameter name {name!r}")
    return m["base"].rstrip(), m["tag"]


def _combine_name(base: str, idx: int, tag: str | None = None) -> str:
    """Return ``'base (idx)'`` or ``'base (idx) [tag]'``."""
    if idx < 0:
        raise ValueError("idx must be ≥ 0")
    out = f"{base} ({idx})"
    return _add_units(out, tag)


def _add_units(base: str, tag: str | None) -> str:
    """Return ``'base'`` or ``'base [tag]'``."""
    out = f"{base}"
    if tag:
        out += f" [{tag}]"
    return out


def arrayize_dict(
    scalar_dict: dict[str, Any],
) -> dict[str, Any]:
    """
    Collapse scalar keys back into lists.  Tags are kept with the base name
    (``'a [V]'``).  A sequence is collapsed only if indices 0…N are all present.

    Parameters
    ----------
    scalar_dict : dict[str, Any]
        The dictionary to arrayize

    Returns
    -------
    dict[str, Any]
        The arrayized dictionary
    """
    out = {}
    processed = set()

    # discover (base, tag) pairs that appear scalarised
    pairs = set()
    for k in scalar_dict:
        match = _KeyMatch(k)
        if match:
            pairs.add((match.base, match.tag))

    # rebuild each pair
    for base, tag in pairs:
        idx_val = {}
        own_keys = []

        for k, v in scalar_dict.items():
            match = _KeyMatch(k)
            if match and match.base == base and match.tag == tag:
                if match.idx in idx_val:
                    raise ValueError(f"Duplicate index {match.idx} for '{base}'")
                idx_val[match.idx] = v
                own_keys.append(k)

        if not idx_val:
            raise ValueError(
                f"No indices found for '{base}'. "
                "It should not be possible to reach here. Please report this bug."
            )

        indices = set(idx_val)
        if not _contiguous_and_ordered_indices(indices):
            missing = sorted(set(range(max(indices) + 1)) - indices)
            raise ValueError(f"Missing indices {missing} for '{base}'")

        collapsed_key = _add_units(base, tag)
        if collapsed_key in out:
            raise ValueError(f"Duplicate key after rebuild: {collapsed_key!r}")

        out[collapsed_key] = [idx_val[i] for i in range(max(idx_val) + 1)]
        processed.update(own_keys)

    # copy untouched scalars
    for k, v in scalar_dict.items():
        if k not in processed:
            if k in out:
                raise ValueError(f"Duplicate key: {k!r}")
            out[k] = v

    return out


def _contiguous_and_ordered_indices(indices: set[int]) -> bool:
    """
    Check if a set of indices forms a contiguous sequence starting from 0
    to max(indices).

    Parameters
    ----------
    indices : set[int]
        Set of integer indices to check for contiguity

    Returns
    -------
    bool
        True if indices form the sequence {0, 1, 2, ..., max(indices)},
        False otherwise. Returns False for empty sets.
    """
    return bool(indices) and indices == set(range(max(indices) + 1))


def convert_parameter_values_to_json(parameter_values, filename=None):
    """
    Converts a ParameterValues object to a JSON-serializable dictionary and optionally
    saves it to a file.

    Parameters
    ----------
    parameter_values : ParameterValues
        The ParameterValues object to convert

    filename : str, optional
        The filename to save the JSON file to. If not provided, the dictionary is
        not saved.
    """
    parameter_values_dict = {}

    for k, v in parameter_values.items():
        if callable(v):
            parameter_values_dict[k] = convert_symbol_to_json(
                convert_function_to_symbolic_expression(v, k)
            )
        else:
            parameter_values_dict[k] = convert_symbol_to_json(v)

    if filename is not None:
        with open(filename, "w") as f:
            json.dump(
                parameter_values_dict, f, indent=2, default=Serialise._json_encoder
            )

    return parameter_values_dict



#
# Class for SEI growth
#
import pybamm

from .base_sei import BaseModel


class SEIGrowth(BaseModel):
    """
    Class for SEI growth.

    Most of the models are from Section 5.6.4 of :footcite:t:`Marquis2020` and
    references therein.

    The ec reaction limited model is from :footcite:t:`Yang2017`.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reaction_loc : str
        Where the reaction happens: "x-average" (SPM, SPMe, etc),
        "full electrode" (full DFN), or "interface" (half-cell model)
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    cracks : bool, optional
        Whether this is a submodel for standard SEI or SEI on cracks
    """

    def __init__(self, param, domain, options, phase="primary", cracks=False):
        super().__init__(param, domain, options=options, phase=phase, cracks=cracks)
        if self.options.electrode_types[domain] == "planar":
            self.reaction_loc = "interface"
        elif self.options["x-average side reactions"] == "true":
            self.reaction_loc = "x-average"
        else:
            self.reaction_loc = "full electrode"
        SEI_option = getattr(self.options, domain)["SEI"]
        if SEI_option == "ec reaction limited":
            pybamm.citations.register("Yang2017")
        elif SEI_option == "VonKolzenberg2020":
            pybamm.citations.register("VonKolzenberg2020")
        elif SEI_option == "tunnelling limited":
            pybamm.citations.register("Tang2012")
        else:
            pybamm.citations.register("Marquis2020")

    def get_fundamental_variables(self):
        domain, Domain = self.domain_Domain
        L_sei_0 = self.phase_param.L_sei_0
        V_bar_sei = self.phase_param.V_bar_sei
        if self.reaction_loc == "x-average":
            # c_sei_xav and c_sei are bulk quantities [mol.m-3]
            scale = L_sei_0 * self.phase_param.a_typ / V_bar_sei
            c_sei_xav = pybamm.Variable(
                f"X-averaged {domain} {self.reaction_name}concentration [mol.m-3]",
                domain="current collector",
                scale=scale,
            )
            c_sei_xav.print_name = "c_sei_xav"
            c_sei = pybamm.PrimaryBroadcast(c_sei_xav, f"{domain} electrode")
        elif self.reaction_loc == "full electrode":
            # c_sei is a bulk quantity [mol.m-3]
            scale = L_sei_0 * self.phase_param.a_typ / V_bar_sei
            c_sei = pybamm.Variable(
                f"{Domain} {self.reaction_name}concentration [mol.m-3]",
                domain=f"{domain} electrode",
                auxiliary_domains={"secondary": "current collector"},
                scale=scale,
            )
        elif self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            scale = L_sei_0 / V_bar_sei
            c_sei = pybamm.Variable(
                f"{Domain} {self.reaction_name}concentration [mol.m-2]",
                domain="current collector",
                scale=scale,
            )
        c_sei.print_name = "c_sei"

        variables = self._get_standard_concentration_variables(c_sei)

        return variables

    def get_coupled_variables(self, variables):
        phase_param = self.phase_param
        domain, Domain = self.domain_Domain
        SEI_option = getattr(getattr(self.options, domain), self.phase)["SEI"]
        T = variables[f"{Domain} electrode temperature [K]"]
        # delta_phi = phi_s - phi_e
        if self.reaction_loc == "interface":
            delta_phi = variables[
                "Lithium metal interface surface potential difference [V]"
            ]
            T = pybamm.boundary_value(T, "right")
        else:
            delta_phi = variables[
                f"{Domain} electrode surface potential difference [V]"
            ]

        # Look for current that contributes to the -IR drop
        # If we can't find the interfacial current density from the main reaction, j,
        # it's ok to fall back on the total interfacial current density, j_tot
        # This should only happen when the interface submodel is "InverseButlerVolmer"
        # in which case j = j_tot (uniform) anyway
        if f"{Domain} electrode interfacial current density [A.m-2]" in variables:
            j = variables[f"{Domain} electrode interfacial current density [A.m-2]"]
        elif self.reaction_loc == "interface":
            j = variables["Lithium metal total interfacial current density [A.m-2]"]
        else:
            j = variables[
                f"X-averaged {domain} electrode total "
                "interfacial current density [A.m-2]"
            ]

        L_sei = variables[f"{Domain} {self.reaction_name}thickness [m]"]

        R_sei = phase_param.R_sei
        eta_SEI = delta_phi - phase_param.U_sei - j * L_sei * R_sei
        # Thermal prefactor for reaction, interstitial and EC models
        F_RT = self.param.F / (self.param.R * T)

        # Define alpha_SEI depending on whether it is symmetric or asymmetric. This
        # applies to "reaction limited" and "EC reaction limited"
        if SEI_option.endswith("(asymmetric)"):
            alpha_SEI = phase_param.alpha_SEI
        else:
            alpha_SEI = 0.5

        if SEI_option.startswith("reaction limited"):
            # Scott Marquis thesis (eq. 5.92)
            j_sei = -phase_param.j0_sei * pybamm.exp(-alpha_SEI * F_RT * eta_SEI)

        elif SEI_option == "electron-migration limited":
            # Scott Marquis thesis (eq. 5.94)
            eta_inner = delta_phi - phase_param.U_sei
            j_sei = (eta_inner < 0) * phase_param.kappa_inner * eta_inner / L_sei

        elif SEI_option == "tunnelling limited":  #
            # This comes from eq.25 in Tang, M., Lu, S. and Newman, J., 2012.
            # Experimental and theoretical investigation of solid-electrolyte-interphase
            # formation mechanisms on glassy carbon.
            # Journal of The Electrochemical Society, 159(11), p.A1775.
            j_sei = (
                -phase_param.j0_sei
                * pybamm.exp(-alpha_SEI * F_RT * delta_phi)
                * pybamm.exp(-phase_param.beta_tunnelling * L_sei)
            )

        elif SEI_option == "VonKolzenberg2020":
            # Equation 19 in
            # von Kolzenberg L, Latz A, Horstmann B.
            # Solid electrolyte interphase during battery cycling:
            # Theory of growth regimes. ChemSusChem. 2020 Aug 7;13(15):3901-10.
            eta_bar = F_RT * (delta_phi)
            L_diff = (
                phase_param.D_li
                * phase_param.c_li_0
                * self.param.F
                / phase_param.j0_sei
            ) * pybamm.exp(-(1 - alpha_SEI) * eta_bar)
            L_tun = phase_param.L_tunneling
            L_app = (L_sei - L_tun) * ((L_sei - L_tun) > 0)
            L_mig = (
                2
                / F_RT
                * phase_param.kappa_Li_ion
                / pybamm.maximum(pybamm.AbsoluteValue(j), 1e-38)
            )
            sign_j = 2 * (j > 0) - 1
            LL_k = (1 - L_app / L_mig * sign_j) / (
                1 - L_app / L_mig * sign_j + L_app / L_diff
            )
            j_sei = -phase_param.j0_sei * LL_k * pybamm.exp(-alpha_SEI * eta_bar)

        elif SEI_option == "interstitial-diffusion limited":
            # Scott Marquis thesis (eq. 5.96)
            j_sei = -(
                phase_param.D_li * phase_param.c_li_0 * self.param.F / L_sei
            ) * pybamm.exp(-F_RT * delta_phi)

        elif SEI_option == "solvent-diffusion limited":
            # Scott Marquis thesis (eq. 5.91)
            j_sei = -phase_param.D_sol * phase_param.c_sol * self.param.F / L_sei

        elif SEI_option.startswith("ec reaction limited"):
            # we have a linear system for j and c
            #  c = c_0 + j * L / F / D          [1] (eq 11 in the Yang2017 paper)
            #  j = - F * c * k_exp()            [2] (eq 10 in the Yang2017 paper, factor
            #                                        of a is outside the defn of j here)
            # [1] into [2] gives (F cancels in the second terms)
            #  j = - F * c_0 * k_exp() - j * L * k_exp() / D
            # rearrange
            #  j = -F * c_0* k_exp() / (1 + L * k_exp() / D)
            #  c_ec = c_0 - L * k_exp() / D / (1 + L * k_exp() / D)
            #       = c_0 / (1 + L * k_exp() / D)
            k_exp = phase_param.k_sei * pybamm.exp(-alpha_SEI * F_RT * eta_SEI)
            L_over_D = L_sei / phase_param.D_ec
            c_0 = phase_param.c_ec_0
            j_sei = -self.param.F * c_0 * k_exp / (1 + L_over_D * k_exp)
            c_ec = c_0 / (1 + L_over_D * k_exp)

            # Get variables related to the concentration
            c_ec_av = pybamm.x_average(c_ec)

            if self.reaction == "SEI on cracks":
                name = f"{Domain} EC concentration on cracks [mol.m-3]"
            else:
                name = f"{Domain} EC surface concentration [mol.m-3]"
            variables.update({name: c_ec, f"X-averaged {name}": c_ec_av})

        # All SEI growth mechanisms assumed to have Arrhenius dependence
        arrhenius = pybamm.exp(
            phase_param.E_sei / self.param.R * (1 / self.param.T_ref - 1 / T)
        )

        j_sei = arrhenius * j_sei

        variables.update(self._get_standard_reaction_variables(j_sei))

        # Add other standard coupled variables
        variables.update(super().get_coupled_variables(variables))

        return variables

    def set_rhs(self, variables):
        domain, Domain = self.domain_Domain

        if self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            c_sei = variables[f"{Domain} {self.reaction_name}concentration [mol.m-2]"]
            j_sei = variables[
                f"{Domain} electrode {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]
            a = 1
        elif self.reaction_loc == "x-average":
            # c_sei is a bulk quanitity [mol.m-3]
            c_sei = variables[
                f"X-averaged {domain} {self.reaction_name}concentration [mol.m-3]"
            ]
            j_sei = variables[
                f"X-averaged {domain} electrode {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]
            a = variables[
                f"X-averaged {domain} electrode {self.phase_name}"
                "surface area to volume ratio [m-1]"
            ]
        else:
            c_sei = variables[f"{Domain} {self.reaction_name}concentration [mol.m-3]"]
            j_sei = variables[
                f"{Domain} electrode {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]
            a = variables[
                f"{Domain} electrode {self.phase_name}"
                "surface area to volume ratio [m-1]"
            ]

        # For SEI on cracks, need to use crack area instead of surface area
        # To do this, use the roughness parameter, which works as follows:
        # a + a_cr = roughness * a
        # Rearrange to get a_cr = (roughness - 1) * a
        # This is done here by replacing a with a_cr using a *= roughness - 1
        if self.reaction == "SEI on cracks":
            if self.reaction_loc == "x-average":
                roughness = variables[f"X-averaged {domain} electrode roughness ratio"]
            else:
                roughness = variables[f"{Domain} electrode roughness ratio"]
            a *= roughness - 1  # Replace surface area with crack area

        # a * j_sei / F is the rate of consumption of Li moles by SEI reaction
        # 1/z_sei converts from Li moles to SEI moles (z_sei=Li mol per sei mol)
        # a * j_sei / (F * z_sei) = rate of consumption of SEI moles by SEI reaction
        dcdt_sei = a * j_sei / (self.param.F * self.phase_param.z_sei)
        # Therefore, -a * j_sei / (F * z_sei) = rate of creation of SEI moles
        self.rhs = {c_sei: -dcdt_sei}

    def set_initial_conditions(self, variables):
        domain, Domain = self.domain_Domain
        L_sei_0 = self.phase_param.L_sei_0
        V_bar_sei = self.phase_param.V_bar_sei
        if self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            c_sei = variables[f"{Domain} {self.reaction_name}concentration [mol.m-2]"]
            c_sei_0 = L_sei_0 / V_bar_sei
        elif self.reaction_loc == "x-average":
            # c_sei is a bulk quantity [mol.m-3]
            c_sei = variables[
                f"X-averaged {domain} {self.reaction_name}concentration [mol.m-3]"
            ]
            c_sei_0 = L_sei_0 * self.phase_param.a_typ / V_bar_sei
            c_sei_cr0 = self.phase_param.L_sei_cr0 * self.phase_param.a_typ / V_bar_sei
        else:
            # c_sei is a bulk quantity [mol.m-3]
            c_sei = variables[f"{Domain} {self.reaction_name}concentration [mol.m-3]"]
            c_sei_0 = L_sei_0 * self.phase_param.a_typ / V_bar_sei
            c_sei_cr0 = self.phase_param.L_sei_cr0 * self.phase_param.a_typ / V_bar_sei

        if self.reaction == "SEI on cracks":
            self.initial_conditions = {c_sei: c_sei_cr0}
        else:
            self.initial_conditions = {c_sei: c_sei_0}



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import casadi
import numpy as np
import numpy.typing as npt
from scipy.integrate import trapezoid

import pybamm


@dataclass
class ProcessedVariableTimeIntegral:
    method: Literal["discrete", "continuous"]
    sum_node: pybamm.Symbol
    initial_condition: npt.NDArray[np.float64] | float
    discrete_times: npt.NDArray[np.float64] | None
    post_sum_node: pybamm.Symbol | None = None
    post_sum: casadi.Function | None = None

    def postfix_sum(self, entries, t_pts) -> np.ndarray:
        if self.method == "discrete":
            return np.sum(
                entries, axis=0, initial=self.initial_condition, keepdims=True
            )
        else:
            return np.array(
                [trapezoid(entries, t_pts, axis=0) + float(self.initial_condition)]
            )

    def postfix(self, entries, t_pts, inputs) -> np.ndarray:
        """
        Compute the postfix sum or integral of the entries over time points,
        """
        the_integral = self.postfix_sum(entries, t_pts)
        if self.post_sum_node is None:
            ret = the_integral
        elif self.post_sum is None:
            ret = self.post_sum_node.evaluate(0.0, the_integral, None, inputs).reshape(
                -1
            )
        else:
            ret = self.post_sum(0.0, the_integral, inputs).full().reshape(-1)
        return ret

    def postfix_sensitivities(
        self,
        var_name,
        entries,
        t_pts,
        inputs,
        sensitivities,
    ) -> np.ndarray:
        # post fix for discrete time integral won't give correct result
        # if ts are not equal to the discrete times. Raise error
        # in this case
        if self.method == "discrete":
            if not (
                len(t_pts) == len(self.discrete_times)
                and np.allclose(t_pts, self.discrete_times, atol=1e-10)
            ):
                raise pybamm.SolverError(
                    f'Processing discrete-time-sum variable "{var_name}": solution times '
                    "and discrete times of the time integral are not equal. Set 't_interp=discrete_sum_times' to "
                    f"ensure the correct times are used.\nSolution times: {t_pts}\nDiscrete Sum times: {self.discrete_times}"
                )

        the_integral = self.postfix_sum(sensitivities, t_pts)
        if self.post_sum_node is None:
            return the_integral

        y_casadi = casadi.MX.sym("y", entries.shape[0])
        sens_casadi = casadi.MX.sym("s_var", the_integral.shape)
        t_casadi = casadi.MX.sym("t")
        p_casadi = {
            name: casadi.MX.sym(
                name, 1 if not isinstance(value, np.ndarray) else value.shape[0]
            )
            for name, value in inputs.items()
        }
        p_casadi_stacked = casadi.vertcat(*[p for p in p_casadi.values()])
        inputs_stacked = casadi.vertcat(*[v for v in inputs.values()])
        post_sum_casadi = self.post_sum_node.to_casadi(
            t_casadi, y_casadi, inputs=p_casadi
        )

        dpost_dy = casadi.jacobian(post_sum_casadi, y_casadi)
        dpost_dp = casadi.jacobian(post_sum_casadi, p_casadi_stacked)
        sens = dpost_dy @ sens_casadi + dpost_dp
        sens_fun = casadi.Function(
            "sens_fun",
            [t_casadi, y_casadi, p_casadi_stacked, sens_casadi],
            [sens],
        )
        sens_values = sens_fun(0.0, entries, inputs_stacked, the_integral)
        return sens_values.full()

    @staticmethod
    def to_post_sum_expr(
        var: pybamm.Symbol,
        integral_node: pybamm.ExplicitTimeIntegral | pybamm.DiscreteTimeSum,
        nstates: int,
    ) -> pybamm.Symbol:
        if var == integral_node:
            # If the variable is the integral node itself, return a state vector
            return pybamm.StateVector(slice(0, nstates))

        # Should not be any time or state vector nodes in the children
        if any(isinstance(c, pybamm.Time | pybamm.StateVector) for c in var.children):
            raise ValueError(
                "For expressions containing time integrals, "
                "time or state vector nodes should only appear within the "
                "time integral node. "
            )
        new_children = [
            ProcessedVariableTimeIntegral.to_post_sum_expr(c, integral_node, nstates)
            for c in var.children
        ]
        return var.create_copy(new_children)

    @staticmethod
    def from_pybamm_var(
        var: pybamm.Symbol,
        nstates: int,
    ) -> ProcessedVariableTimeIntegral | None:
        sum_node = None
        for symbol in var.pre_order():
            if isinstance(symbol, pybamm.ExplicitTimeIntegral | pybamm.DiscreteTimeSum):
                if sum_node is None:
                    sum_node = symbol
                else:
                    raise ValueError(
                        "More than one time integral node found in the variable"
                        f" {var.name}. Only one time integral node is allowed."
                    )
        if sum_node is None:
            return None
        if sum_node == var:
            post_sum_node = None
        else:
            sum_y_len = sum_node.evaluate_for_shape().shape[0]
            post_sum_node = ProcessedVariableTimeIntegral.to_post_sum_expr(
                var, sum_node, sum_y_len
            )
        if isinstance(sum_node, pybamm.DiscreteTimeSum):
            return ProcessedVariableTimeIntegral(
                method="discrete",
                post_sum_node=post_sum_node,
                sum_node=sum_node,
                initial_condition=0.0,
                discrete_times=sum_node.sum_times,
            )
        elif isinstance(sum_node, pybamm.ExplicitTimeIntegral):
            if isinstance(sum_node.initial_condition, pybamm.Symbol):
                initial_condition = sum_node.initial_condition.evaluate()
            else:
                initial_condition = sum_node.initial_condition
            return ProcessedVariableTimeIntegral(
                method="continuous",
                post_sum_node=post_sum_node,
                sum_node=sum_node,
                initial_condition=initial_condition,
                discrete_times=None,
            )
        else:
            raise ValueError("Unsupported variable type")  # pragma: no cover



#
# A model to calculate electrode-specific SOH, adapted for composite electrodes
#
import pybamm

from .util import _get_lithiation_delithiation, check_if_composite


def _get_stoich_variables(options):
    variables = {
        "x_100_1": pybamm.Variable("x_100_1"),
        "y_100_1": pybamm.Variable("y_100_1"),
        "x_0_1": pybamm.Variable("x_0_1"),
        "y_0_1": pybamm.Variable("y_0_1"),
        "x_init_1": pybamm.Variable("x_init_1"),
        "y_init_1": pybamm.Variable("y_init_1"),
    }
    is_positive_composite = check_if_composite(options, "positive")
    is_negative_composite = check_if_composite(options, "negative")
    if is_positive_composite:
        variables["y_100_2"] = pybamm.Variable("y_100_2")
        variables["y_0_2"] = pybamm.Variable("y_0_2")
        variables["y_init_2"] = pybamm.Variable("y_init_2")
    if is_negative_composite:
        variables["x_100_2"] = pybamm.Variable("x_100_2")
        variables["x_0_2"] = pybamm.Variable("x_0_2")
        variables["x_init_2"] = pybamm.Variable("x_init_2")
    return variables


def _get_initial_conditions(options, soc_init):
    variables = _get_stoich_variables(options)
    ics = {}
    for name, var in variables.items():
        if "100" in name and "x" in name:
            ics[var] = 0.8
        elif "0" in name and "x" in name:
            ics[var] = 0.2
        elif "100" in name and "y" in name:
            ics[var] = 0.8
        elif "0" in name and "y" in name:
            ics[var] = 0.2
        elif "init" in name and "x" in name:
            ics[var] = soc_init
        elif "init" in name and "y" in name:
            ics[var] = 1 - soc_init
    return ics


def _get_direction(electrode):
    if electrode == "positive":
        return pybamm.Scalar(-1)
    else:
        return pybamm.Scalar(1)


def _get_prefix(electrode):
    if electrode == "positive":
        return "y"
    else:
        return "x"


def _get_electrode_capacity_equation(options, electrode):
    prefix = _get_prefix(electrode)
    e = electrode[0]
    i_am_composite = check_if_composite(options, electrode)
    stoich_variables = _get_stoich_variables(options)
    direction = _get_direction(electrode)
    Q_1 = pybamm.InputParameter(f"Q_{e}_1")
    Q = (
        direction
        * (stoich_variables[f"{prefix}_100_1"] - stoich_variables[f"{prefix}_0_1"])
        * Q_1
    )
    if i_am_composite:
        Q_2 = pybamm.InputParameter(f"Q_{e}_2")
        Q += (
            direction
            * (stoich_variables[f"{prefix}_100_2"] - stoich_variables[f"{prefix}_0_2"])
            * Q_2
        )
    return Q


def _get_cyclable_lithium_equation(options, soc="100"):
    x_soc_1 = pybamm.Variable(f"x_{soc}_1")
    y_soc_1 = pybamm.Variable(f"y_{soc}_1")
    Q_n_1 = pybamm.InputParameter("Q_n_1")
    Q_p_1 = pybamm.InputParameter("Q_p_1")
    lithium_primary_phases = Q_n_1 * x_soc_1 + Q_p_1 * y_soc_1
    lithium_secondary_phases = 0.0
    is_positive_composite = check_if_composite(options, "positive")
    is_negative_composite = check_if_composite(options, "negative")
    if is_positive_composite:
        Q_p_2 = pybamm.InputParameter("Q_p_2")
        y_soc_2 = pybamm.Variable(f"y_{soc}_2")
        lithium_secondary_phases += Q_p_2 * y_soc_2
    if is_negative_composite:
        Q_n_2 = pybamm.InputParameter("Q_n_2")
        x_soc_2 = pybamm.Variable(f"x_{soc}_2")
        lithium_secondary_phases += Q_n_2 * x_soc_2
    return lithium_primary_phases + lithium_secondary_phases


class ElectrodeSOHComposite(pybamm.BaseModel):
    """Model to calculate electrode-specific SOH for a cell with composite electrodes,
    adapted from :footcite:t:`Mohtat2019`. This model is mainly for internal use, to
    calculate summary variables in a
    simulation.

    Subscript 1 indicates primary phase and subscript 2 indicates secondary phase.
    """

    def __init__(
        self,
        options,
        direction=None,
        name="ElectrodeSOH model",
        initialization_method="voltage",
    ):
        pybamm.citations.register("Mohtat2019")
        super().__init__(name)
        param = pybamm.LithiumIonParameters(options)

        # Start by just assuming known value is cyclable lithium (and solve all at once)
        Q_Li = pybamm.InputParameter("Q_Li")
        is_negative_composite = check_if_composite(options, "negative")
        is_positive_composite = check_if_composite(options, "positive")
        variables = _get_stoich_variables(options)
        x_100_1 = variables["x_100_1"]
        y_100_1 = variables["y_100_1"]
        x_0_1 = variables["x_0_1"]
        y_0_1 = variables["y_0_1"]
        V_max = param.voltage_high_cut
        V_min = param.voltage_low_cut
        # Here we use T_ref as the stoichiometry limits are defined using the reference
        # state
        if is_negative_composite:
            x_100_2 = variables["x_100_2"]
            x_0_2 = variables["x_0_2"]
            self.algebraic[x_100_2] = param.n.sec.U(
                x_100_2,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="secondary"
                ),
            ) - param.n.prim.U(
                x_100_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="primary"
                ),
            )
            self.algebraic[x_0_2] = param.n.sec.U(
                x_0_2,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="secondary"
                ),
            ) - param.n.prim.U(
                x_0_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="primary"
                ),
            )
        if is_positive_composite:
            y_100_2 = variables["y_100_2"]
            y_0_2 = variables["y_0_2"]
            self.algebraic[y_100_2] = param.p.sec.U(
                y_100_2,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="secondary"
                ),
            ) - param.p.prim.U(
                y_100_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="primary"
                ),
            )
            self.algebraic[y_0_2] = param.p.prim.U(
                y_0_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="primary"
                ),
            ) - param.p.sec.U(
                y_0_2,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="secondary"
                ),
            )
        self.algebraic[x_100_1] = (
            param.p.prim.U(
                y_100_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="primary"
                ),
            )
            - param.n.prim.U(
                x_100_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="primary"
                ),
            )
            - V_max
        )
        self.algebraic[x_0_1] = (
            param.p.prim.U(
                y_0_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="primary"
                ),
            )
            - param.n.prim.U(
                x_0_1,
                param.T_ref,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="primary"
                ),
            )
            - V_min
        )
        # arbitrary choice: use y_0_1 for the capacity equation
        self.algebraic[y_0_1] = _get_electrode_capacity_equation(
            options, "positive"
        ) - _get_electrode_capacity_equation(options, "negative")
        self.algebraic[y_100_1] = Q_Li - _get_cyclable_lithium_equation(options)

        x_init_1 = variables["x_init_1"]
        y_init_1 = variables["y_init_1"]
        if initialization_method == "voltage":
            # Here we use T_init so that the initial voltage is correct, including the
            # contribution from the entropic change
            V_init = pybamm.InputParameter("V_init")
            self.algebraic[x_init_1] = (
                param.p.prim.U(
                    y_init_1,
                    param.T_init,
                    _get_lithiation_delithiation(
                        direction, "positive", options, phase="primary"
                    ),
                )
                - param.n.prim.U(
                    x_init_1,
                    param.T_init,
                    _get_lithiation_delithiation(
                        direction, "negative", options, phase="primary"
                    ),
                )
                - V_init
            )
            self.algebraic[y_init_1] = (
                _get_cyclable_lithium_equation(options, "init") - Q_Li
            )
        elif initialization_method == "SOC":
            soc_init = pybamm.InputParameter("SOC_init")
            negative_soc = x_init_1 * pybamm.InputParameter("Q_n_1")
            if is_negative_composite:
                x_init_2 = variables["x_init_2"]
                negative_soc += x_init_2 * pybamm.InputParameter("Q_n_2")

            negative_0_soc = x_0_1 * pybamm.InputParameter("Q_n_1")
            if is_negative_composite:
                negative_0_soc += x_0_2 * pybamm.InputParameter("Q_n_2")

            negative_100_soc = x_100_1 * pybamm.InputParameter("Q_n_1")
            if is_negative_composite:
                negative_100_soc += x_100_2 * pybamm.InputParameter("Q_n_2")
            self.algebraic[x_init_1] = (
                (negative_soc - negative_0_soc) / (negative_100_soc - negative_0_soc)
            ) - soc_init
            self.algebraic[y_init_1] = (
                _get_cyclable_lithium_equation(options, "init") - Q_Li
            )
        else:
            raise ValueError("Invalid initialization method")
        # Add voltage equations for secondary phases (init), we use T_ref if setting
        # based on SOC since the stoichiometry limits are defined using the reference
        # state, and T_init if setting based on voltage since the entropic change is
        # included in the voltage equation
        T = param.T_init if initialization_method == "voltage" else param.T_ref
        if is_positive_composite:
            y_init_2 = variables["y_init_2"]
            self.algebraic[y_init_2] = param.p.prim.U(
                y_init_1,
                T,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="primary"
                ),
            ) - param.p.sec.U(
                y_init_2,
                T,
                _get_lithiation_delithiation(
                    direction, "positive", options, phase="secondary"
                ),
            )
        if is_negative_composite:
            x_init_2 = variables["x_init_2"]
            self.algebraic[x_init_2] = param.n.prim.U(
                x_init_1,
                T,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="primary"
                ),
            ) - param.n.sec.U(
                x_init_2,
                T,
                _get_lithiation_delithiation(
                    direction, "negative", options, phase="secondary"
                ),
            )

        self.variables.update(variables)
        if initialization_method == "SOC":
            soc_init = pybamm.InputParameter("SOC_init")
        else:
            soc_init = 0.5
        self.initial_conditions.update(_get_initial_conditions(options, soc_init))

    @property
    def default_solver(self):
        # Use AlgebraicSolver as CasadiAlgebraicSolver gives unnecessary warnings
        return pybamm.AlgebraicSolver()


def get_initial_stoichiometries_composite(
    initial_value,
    parameter_values,
    direction=None,
    param=None,
    options=None,
    tol=1e-6,
    inputs=None,
    known_value="cyclable lithium capacity",
    **kwargs,
):
    """
    Get the minimum and maximum stoichiometries from the parameter values

        Parameters
        ----------
        initial_value : float
            Target initial value.
            If integer, interpreted as SOC, must be between 0 and 1.
            If string e.g. "4 V", interpreted as voltage,
            must be between V_min and V_max.
        direction : str, optional
            The OCV branch to use in the electrode SOH model. Can be "charge" or
            "discharge".
        tol : float, optional
            The tolerance for the solver used to compute the initial stoichiometries.
            A lower value results in higher precision but may increase computation time.
            Default is 1e-6.
        inputs : dict, optional
            A dictionary of input parameters passed to the model.
        known_value : str, optional
            The known value needed to complete the electrode SOH model.
            Can be "cyclable lithium capacity".
    """
    inputs = inputs or {}
    if known_value != "cyclable lithium capacity":
        raise ValueError(
            "Only `cyclable lithium capacity` is supported for composite electrodes"
        )

    Q_n_1 = parameter_values.evaluate(param.n.prim.Q_init, inputs=inputs)
    Q_p_1 = parameter_values.evaluate(param.p.prim.Q_init, inputs=inputs)
    is_positive_composite = check_if_composite(options, "positive")
    is_negative_composite = check_if_composite(options, "negative")
    Qs = {
        "Q_n_1": Q_n_1,
        "Q_p_1": Q_p_1,
    }
    if is_positive_composite:
        Q_p_2 = parameter_values.evaluate(param.p.sec.Q_init, inputs=inputs)
        Qs["Q_p_2"] = Q_p_2
    if is_negative_composite:
        Q_n_2 = parameter_values.evaluate(param.n.sec.Q_init, inputs=inputs)
        Qs["Q_n_2"] = Q_n_2

    Q_Li = parameter_values.evaluate(param.Q_Li_particles_init, inputs=inputs)
    all_inputs = {**inputs, **Qs, "Q_Li": Q_Li}
    # Solve the model and check outputs
    if isinstance(initial_value, str) and initial_value.endswith("V"):
        all_inputs["V_init"] = float(initial_value[:-1])
        initialization_method = "voltage"
    elif isinstance(initial_value, float) and initial_value >= 0 and initial_value <= 1:
        initialization_method = "SOC"
        all_inputs["SOC_init"] = initial_value
    else:
        raise ValueError("Invalid initial value")
    model = ElectrodeSOHComposite(
        options, direction, initialization_method=initialization_method
    )
    sim = pybamm.Simulation(
        model, parameter_values=parameter_values, solver=pybamm.AlgebraicSolver(tol=tol)
    )
    sol = sim.solve([0, 1], inputs=all_inputs)
    return {var: sol[var].entries[0] for var in model.variables.keys()}



#
# Class for reaction driven porosity changes as an ODE
#
import pybamm

from .base_porosity import BaseModel


class ReactionDrivenODE(BaseModel):
    """Reaction-driven porosity changes as an ODE

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        Options dictionary passed from the full model
    x_average : bool
        Whether to use x-averaged variables (SPM, SPMe, etc) or full variables (DFN)
    """

    def __init__(self, param, options, x_average):
        super().__init__(param, options)
        self.x_average = x_average

    def get_fundamental_variables(self):
        eps_dict = {}
        for domain in self.options.whole_cell_domains:
            Domain = domain.capitalize()
            if self.x_average is True:
                eps_k_av = pybamm.Variable(
                    f"X-averaged {domain} porosity",
                    domain="current collector",
                    bounds=(0, 1),
                )
                eps_k = pybamm.PrimaryBroadcast(eps_k_av, domain)
            else:
                eps_k = pybamm.Variable(
                    f"{Domain} porosity",
                    domain=domain,
                    auxiliary_domains={"secondary": "current collector"},
                    bounds=(0, 1),
                )
            eps_dict[domain] = eps_k
        variables = self._get_standard_porosity_variables(eps_dict)

        return variables

    def get_coupled_variables(self, variables):
        depsdt_dict = {}
        for domain in self.options.whole_cell_domains:
            domain_param = self.param.domain_params[domain.split()[0]]
            if domain == "separator":
                depsdt_k = pybamm.FullBroadcast(0, domain, "current collector")
            else:
                if self.x_average is True:
                    a_j_k_av = variables[
                        f"X-averaged {domain} volumetric "
                        "interfacial current density [A.m-3]"
                    ]
                    depsdt_k_av = domain_param.DeltaVsurf * a_j_k_av / self.param.F
                    depsdt_k = pybamm.PrimaryBroadcast(depsdt_k_av, domain)
                else:
                    Domain = domain.capitalize()
                    a_j_k = variables[
                        f"{Domain} volumetric interfacial current density [A.m-3]"
                    ]
                    depsdt_k = domain_param.DeltaVsurf * a_j_k / self.param.F

            depsdt_dict[domain] = depsdt_k
        variables.update(self._get_standard_porosity_change_variables(depsdt_dict))

        return variables

    def set_rhs(self, variables):
        if self.x_average is True:
            for domain in self.options.whole_cell_domains:
                eps_av = variables[f"X-averaged {domain} porosity"]
                deps_dt_av = variables[f"X-averaged {domain} porosity change [s-1]"]
                self.rhs.update({eps_av: deps_dt_av})
        else:
            eps = variables["Porosity"]
            deps_dt = variables["Porosity change"]
            self.rhs = {eps: deps_dt}

    def set_initial_conditions(self, variables):
        if self.x_average is True:
            for domain in self.options.whole_cell_domains:
                eps_k_av = variables[f"X-averaged {domain} porosity"]
                domain_param = self.param.domain_params[domain.split()[0]]
                self.initial_conditions[eps_k_av] = domain_param.epsilon_init
        else:
            eps = variables["Porosity"]
            self.initial_conditions = {eps: self.param.epsilon_init}

    def add_events_from(self, variables):
        for domain in self.options.whole_cell_domains:
            if domain == "separator":
                continue
            Domain = domain.capitalize()
            eps_k = variables[f"{Domain} porosity"]
            self.events.append(
                pybamm.Event(
                    f"Zero {domain} porosity cut-off",
                    pybamm.min(eps_k),
                    pybamm.EventType.TERMINATION,
                )
            )
            self.events.append(
                pybamm.Event(
                    f"Max {domain} porosity cut-off",
                    1 - pybamm.max(eps_k),
                    pybamm.EventType.TERMINATION,
                )
            )



import pybamm


class VectorField(pybamm.Symbol):
    """
    A node in the expression tree representing a vector field.
    """

    def __init__(self, lr_field, tb_field):
        children = [lr_field, tb_field]
        if lr_field.domain != tb_field.domain:
            raise ValueError("lr_field and tb_field must have the same domain")
        self.lr_field = lr_field
        self.tb_field = tb_field
        super().__init__(name="vector_field", children=children, domain=lr_field.domain)

    def create_copy(
        self,
        new_children: list[pybamm.Symbol] | None = None,
        perform_simplifications: bool = True,
    ):
        if new_children is None:
            new_children = [
                self.lr_field.create_copy(
                    perform_simplifications=perform_simplifications
                ),
                self.tb_field.create_copy(
                    perform_simplifications=perform_simplifications
                ),
            ]
        return VectorField(*new_children)

    def _evaluate_for_shape(self):
        return self.lr_field.evaluate_for_shape()

    def evaluates_on_edges(self, dimension: str) -> bool:
        left_evaluates_on_edges = self.lr_field.evaluates_on_edges(dimension)
        right_evaluates_on_edges = self.tb_field.evaluates_on_edges(dimension)
        if left_evaluates_on_edges:
            left_evaluates_on_edges = True
        if right_evaluates_on_edges:
            right_evaluates_on_edges = True
        if left_evaluates_on_edges == right_evaluates_on_edges:
            return left_evaluates_on_edges
        else:
            raise ValueError(
                "lr_field and tb_field must either both evaluate on edges or both not evaluate on edges"
            )



from .base_ocp import BaseOpenCircuitPotential
from .base_hysteresis_ocp import BaseHysteresisOpenCircuitPotential
from .single_ocp import SingleOpenCircuitPotential
from .current_sigmoid_ocp import CurrentSigmoidOpenCircuitPotential
from .msmr_ocp import MSMROpenCircuitPotential
from .one_state_differential_capacity_hysteresis_ocp import (
    OneStateDifferentialCapacityHysteresisOpenCircuitPotential,
)
from .one_state_hysteresis_ocp import OneStateHysteresisOpenCircuitPotential

__all__ = [
    "base_ocp",
    "base_hysteresis_ocp",
    "current_sigmoid_ocp",
    "msmr_ocp",
    "single_ocp",
    "one_state_differential_capacity_hysteresis_ocp",
    "one_state_hysteresis_ocp",
]



#
# Two-dimensional submeshes
#
import numpy as np

import pybamm

from .meshes import SubMesh


class SubMesh2D(SubMesh):
    """
    2D submesh class.
    Contains the position of the nodes, the number of mesh points, and
    (optionally) information about the tab locations.

    Parameters
    ----------
    edges_lr : array_like
        An array containing the points corresponding to the edges of the submesh
        in the left-right direction
    edges_tb : array_like
        An array containing the points corresponding to the edges of the submesh
        in the top-bottom direction
    coord_sys_lr : string
        The coordinate system of the submesh in the left-right direction
    coord_sys_tb : string
        The coordinate system of the submesh in the top-bottom direction
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs
    """

    def __init__(self, edges_lr, edges_tb, coord_sys, tabs=None):
        self.edges_lr = edges_lr
        self.edges_tb = edges_tb
        self.nodes_lr = (self.edges_lr[1:] + self.edges_lr[:-1]) / 2
        self.nodes_tb = (self.edges_tb[1:] + self.edges_tb[:-1]) / 2
        self.d_edges_lr = np.diff(self.edges_lr)
        self.d_edges_tb = np.diff(self.edges_tb)
        self.d_nodes_lr = np.diff(self.nodes_lr)
        self.d_nodes_tb = np.diff(self.nodes_tb)
        self.npts_lr = self.nodes_lr.size
        self.npts_tb = self.nodes_tb.size
        self.npts = self.npts_lr * self.npts_tb
        self.coord_sys = coord_sys
        self.internal_boundaries = []
        self.dimension = 2

        # Add tab locations in terms of "left" and "right"
        self.tabs = tabs

    def read_lims(self, lims):
        # Separate limits and tabs
        # Read and remove tabs. If "tabs" is not a key in "lims", then tabs is set to
        # "None" and nothing is removed from lims
        tabs = lims.pop("tabs", None)

        # check that only one variable passed in
        if len(lims) != 2:
            raise pybamm.GeometryError("lims should only contain two variables")

        ((spatial_var_lr, spatial_lims_lr), (spatial_var_tb, spatial_lims_tb)) = (
            lims.items()
        )

        if isinstance(spatial_var_lr, str):
            spatial_var_lr = getattr(pybamm.standard_spatial_vars, spatial_var_lr)
        if isinstance(spatial_var_tb, str):
            spatial_var_tb = getattr(pybamm.standard_spatial_vars, spatial_var_tb)

        return spatial_var_lr, spatial_lims_lr, spatial_var_tb, spatial_lims_tb, tabs

    def to_json(self):
        json_dict = {
            "edges_lr": self.edges_lr.tolist(),
            "edges_tb": self.edges_tb.tolist(),
            "coord_sys": self.coord_sys,
        }

        if hasattr(self, "tabs"):
            json_dict["tabs"] = self.tabs

        return json_dict

    def create_ghost_cell(self, side):
        if side == "left":
            gs_edges_lr = np.array(
                [2 * self.edges_lr[0] - self.edges_lr[1], self.edges_lr[0]]
            )
            gs_edges_tb = self.edges_tb
        elif side == "right":
            gs_edges_lr = np.array(
                [self.edges_lr[-1], 2 * self.edges_lr[-1] - self.edges_lr[-2]]
            )
            gs_edges_tb = self.edges_tb
        elif side == "top":
            gs_edges_lr = self.edges_lr
            gs_edges_tb = np.array(
                [2 * self.edges_tb[0] - self.edges_tb[1], self.edges_tb[0]]
            )
        elif side == "bottom":
            gs_edges_lr = self.edges_lr
            gs_edges_tb = np.array(
                [self.edges_tb[-1], 2 * self.edges_tb[-1] - self.edges_tb[-2]]
            )
        else:
            raise ValueError(f"Invalid side: {side}")
        gs_submesh = pybamm.SubMesh2D(gs_edges_lr, gs_edges_tb, self.coord_sys)
        return gs_submesh


class Uniform2DSubMesh(SubMesh2D):
    """
    A 2D submesh with uniform spacing in both dimensions
    """

    def __init__(self, lims, npts):
        spatial_var_lr, spatial_lims_lr, spatial_var_tb, spatial_lims_tb, tabs = (
            self.read_lims(lims)
        )
        npts_lr = npts[spatial_var_lr.name]
        npts_tb = npts[spatial_var_tb.name]

        edges_lr = np.linspace(
            spatial_lims_lr["min"], spatial_lims_lr["max"], npts_lr + 1
        )
        edges_tb = np.linspace(
            spatial_lims_tb["min"], spatial_lims_tb["max"], npts_tb + 1
        )
        if spatial_var_lr.coord_sys != spatial_var_tb.coord_sys:
            raise pybamm.GeometryError(
                "Coordinate systems must be the same for 2D submeshes"
            )
        coord_sys = spatial_var_lr.coord_sys

        super().__init__(edges_lr, edges_tb, coord_sys, tabs=tabs)

    def read_lims(self, lims):
        # Separate limits and tabs
        # Read and remove tabs. If "tabs" is not a key in "lims", then tabs is set to
        # "None" and nothing is removed from lims
        tabs = lims.pop("tabs", None)

        # check that only one variable passed in
        if len(lims) != 2:
            raise pybamm.GeometryError("lims should only contain two variables")

        ((spatial_var1, spatial_lims1), (spatial_var2, spatial_lims2)) = lims.items()

        if isinstance(spatial_var1, str):
            spatial_var1 = getattr(pybamm.standard_spatial_vars, spatial_var1)
        if isinstance(spatial_var2, str):
            spatial_var2 = getattr(pybamm.standard_spatial_vars, spatial_var2)

        return spatial_var1, spatial_lims1, spatial_var2, spatial_lims2, tabs



#
# Class for one-dimensional (x-direction) thermal submodel
#
import pybamm
from pybamm.models.submodels.thermal.base_thermal import BaseThermal


class OneDimensionalX(BaseThermal):
    """
    Class for one-dimensional (x-direction) thermal submodel.
    Note: this model assumes infinitely large electrical and thermal conductivity
    in the current collectors, so that the contribution to the Ohmic heating
    from the current collectors is zero and the boundary conditions are applied
    at the edges of the electrodes (at x=0 and x=1, in non-dimensional coordinates).
    For more information see :footcite:t:`Timms2021` and :footcite:t:`Marquis2020`.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.

    """

    def __init__(self, param, options=None, x_average=False):
        super().__init__(param, options=options, x_average=x_average)
        pybamm.citations.register("Timms2021")

    def get_fundamental_variables(self):
        T_dict = {}
        for domain in ["negative electrode", "separator", "positive electrode"]:
            Domain = domain.capitalize()
            T_k = pybamm.Variable(
                f"{Domain} temperature [K]",
                domain=domain,
                auxiliary_domains={"secondary": "current collector"},
                scale=self.param.T_ref,
            )
            T_dict[domain] = T_k

        T = pybamm.concatenation(*T_dict.values())
        T_cn = pybamm.Variable(
            "Negative current collector temperature [K]",
            domain="current collector",
            scale=self.param.T_ref,
        )
        T_cp = pybamm.Variable(
            "Positive current collector temperature [K]",
            domain="current collector",
            scale=self.param.T_ref,
        )
        T_x_av = self._x_average(T, T_cn, T_cp)
        T_vol_av = self._yz_average(T_x_av)
        T_dict.update(
            {
                "negative current collector": T_cn,
                "positive current collector": T_cp,
                "x-averaged cell": T_x_av,
                "volume-averaged cell": T_vol_av,
            }
        )

        variables = self._get_standard_fundamental_variables(T_dict)
        return variables

    def get_coupled_variables(self, variables):
        variables.update(self._get_standard_coupled_variables(variables))
        return variables

    def set_rhs(self, variables):
        T = variables["Cell temperature [K]"]
        T_cn = variables["Negative current collector temperature [K]"]
        T_n = variables["Negative electrode temperature [K]"]
        T_s = variables["Separator temperature [K]"]
        T_p = variables["Positive electrode temperature [K]"]
        T_cp = variables["Positive current collector temperature [K]"]
        Q = variables["Total heating [W.m-3]"]
        Q_cn = variables["Negative current collector Ohmic heating [W.m-3]"]
        Q_cp = variables["Positive current collector Ohmic heating [W.m-3]"]
        T_surf = variables["Surface temperature [K]"]
        y = pybamm.standard_spatial_vars.y
        z = pybamm.standard_spatial_vars.z

        # Define volumetric heat capacity for electrode/separator/electrode sandwich
        rho_c_p = pybamm.concatenation(
            self.param.n.rho_c_p(T_n),
            self.param.s.rho_c_p(T_s),
            self.param.p.rho_c_p(T_p),
        )

        # Define thermal conductivity for electrode/separator/electrode sandwich
        lambda_ = pybamm.concatenation(
            self.param.n.lambda_(T_n),
            self.param.s.lambda_(T_s),
            self.param.p.lambda_(T_p),
        )

        # Calculate edge/tab cooling
        L_y = self.param.L_y
        L_z = self.param.L_z
        L_cn = self.param.n.L_cc
        L_cp = self.param.p.L_cc
        h_cn = self.param.n.h_cc(y, z)
        h_cp = self.param.p.h_cc(y, z)
        lambda_n = self.param.n.lambda_(T_n)
        lambda_p = self.param.p.lambda_(T_p)
        h_edge = self.param.h_edge(y, z)
        # Negative current collector
        volume_cn = L_cn * L_y * L_z
        negative_tab_area = self.param.n.L_tab * self.param.n.L_cc
        edge_area_cn = 2 * (L_y + L_z) * L_cn - negative_tab_area
        negative_tab_cooling_coefficient = (
            -self.param.n.h_tab * negative_tab_area / volume_cn
        )
        edge_cooling_coefficient_cn = -h_edge * edge_area_cn / volume_cn
        cooling_coefficient_cn = (
            negative_tab_cooling_coefficient + edge_cooling_coefficient_cn
        )
        # Electrode/separator/electrode sandwich
        area_to_volume = (
            2 * (self.param.L_y + self.param.L_z) / (self.param.L_y * self.param.L_z)
        )
        cooling_coefficient = -h_edge * area_to_volume
        # Positive current collector
        volume_cp = L_cp * L_y * L_z
        positive_tab_area = self.param.p.L_tab * self.param.p.L_cc
        edge_area_cp = 2 * (L_y + L_z) * L_cp - positive_tab_area
        positive_tab_cooling_coefficient = (
            -self.param.p.h_tab * positive_tab_area / volume_cp
        )
        edge_cooling_coefficient_cp = -h_edge * edge_area_cp / volume_cp
        cooling_coefficient_cp = (
            positive_tab_cooling_coefficient + edge_cooling_coefficient_cp
        )

        self.rhs = {
            T_cn: (
                (
                    pybamm.boundary_value(lambda_n, "left")
                    * pybamm.boundary_gradient(T_n, "left")
                    - h_cn * (T_cn - T_surf)
                )
                / L_cn
                + Q_cn
                + cooling_coefficient_cn * (T_cn - T_surf)
            )
            / self.param.n.rho_c_p_cc(T_cn),
            T: (
                pybamm.div(lambda_ * pybamm.grad(T))
                + Q
                + cooling_coefficient * (T - T_surf)
            )
            / rho_c_p,
            T_cp: (
                (
                    -pybamm.boundary_value(lambda_p, "right")
                    * pybamm.boundary_gradient(T_p, "right")
                    - h_cp * (T_cp - T_surf)
                )
                / L_cp
                + Q_cp
                + cooling_coefficient_cp * (T_cp - T_surf)
            )
            / self.param.p.rho_c_p_cc(T_cp),
        }

    def set_boundary_conditions(self, variables):
        T = variables["Cell temperature [K]"]
        T_cn = variables["Negative current collector temperature [K]"]
        T_cp = variables["Positive current collector temperature [K]"]

        self.boundary_conditions = {
            T: {"left": (T_cn, "Dirichlet"), "right": (T_cp, "Dirichlet")}
        }

    def set_initial_conditions(self, variables):
        T = variables["Cell temperature [K]"]
        T_cn = variables["Negative current collector temperature [K]"]
        T_cp = variables["Positive current collector temperature [K]"]
        T_init = self.param.T_init
        self.initial_conditions = {
            T_cn: pybamm.boundary_value(T_init, "left"),
            T: T_init,
            T_cp: pybamm.boundary_value(T_init, "right"),
        }
